# -*- coding: utf-8 -*-
#
# pynag - Python Nagios plug-in and configuration environment
# Copyright (C) 2011 Pall Sigurdsson
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""
This module provides a high level Object-Oriented wrapper
around pynag.Parsers.config.

example usage:

from pynag.Parsers import Service,Host

all_services = Service.objects.all
my_service=all_service[0]
print my_service.host_name

example_host = Host.objects.filter(host_name="host.example.com")
canadian_hosts = Host.objects.filter(host_name__endswith=".ca")

for i in canadian_hosts:
    i.alias = "this host is located in Canada"
    i.save()
"""


import os
import re
import subprocess

from pynag import Parsers
from macros import _standard_macros
import all_attributes

# Path To Nagios configuration file
cfg_file = None  # '/etc/nagios/nagios.cfg'

# Were new objects are written by default
pynag_directory = None

# This will be a Parsers.config instance once we have parsed 
config = None 
#config.parse()


#: eventhandlers -- A list of Model.EventHandelers object.
# Event handler is responsible for passing notification whenever something
# important happens in the model.
#
# For example FileLogger class is an event handler responsible for logging to
# file whenever something has been written.
eventhandlers = []


try:
    from collections import defaultdict
except ImportError:
    class defaultdict(dict):
        def __init__(self, default_factory=None, *a, **kw):
            if (default_factory is not None and
                not hasattr(default_factory, '__call__')):
                raise TypeError('first argument must be callable')
            dict.__init__(self, *a, **kw)
            self.default_factory = default_factory
        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                return self.__missing__(key)
        def __missing__(self, key):
            if self.default_factory is None:
                raise KeyError(key)
            self[key] = value = self.default_factory()
            return value
        def __reduce__(self):
            if self.default_factory is None:
                args = tuple()
            else:
                args = self.default_factory,
            return type(self), args, None, None, self.items()
        def copy(self):
            return self.__copy__()
        def __copy__(self):
            return type(self)(self.default_factory, self)
        def __deepcopy__(self, memo):
            import copy
            return type(self)(self.default_factory,
                copy.deepcopy(self.items()))
        def __repr__(self):
            return 'defaultdict(%s, %s)' % (self.default_factory,
                                            dict.__repr__(self))




def debug(text):
    if debug == True: print text


def contains(str1, str2):
    """Returns True if str1 contains str2"""
    if not str1:
        return False
    if str1.find(str2) > -1: return True


def not_contains(str1, str2):
    """Returns True if str1 does not contain str2"""
    return not contains(str1, str2)


def has_field(str1, str2):
    """Returns True if str2 is a field in str1

    For this purpose the string 'example' is a field in '+this,is,an,example'
    """
    str1 = str1.strip('+')
    str1 = str1.split(',')
    str1 = map(lambda x: x.strip(), str1)
    if str2 in str1: return True
    return False


class ObjectRelations(object):
    """ Static container for objects and their respective neighbours """
    # c['contact_name'] = [host1.get_id(),host2.get_id()]
    contact_hosts = defaultdict(set)

    # c['contact_name'] = ['contactgroup1','contactgroup2']
    contact_contactgroups = defaultdict(set)

    # c['contact_name'] = ['service1.get_id()','service2.get_id()']
    contact_services = defaultdict(set)

    # c['contactgroup_name'] = ['contact1.contact_name','contact2.contact_name','contact3.contact_name']
    contactgroup_contacts = defaultdict(set)

    # c['contactgroup_name'] = ['contactgroup1','contactgroup2','contactgroup3']
    contactgroup_contactgroups = defaultdict(set)

    # c['contactgroup_name'] = ['host1.get_id()', 'host2.get_id()']
    contactgroup_hosts = defaultdict(set)

    # c['contactgroup_name'] = ['service1.get_id()', 'service2.get_id()']
    contactgroup_services = defaultdict(set)

    # c['host_name'] = [service1.id, service2.id,service3.id]
    host_services = defaultdict(set)

    # c['host_name'] = ['contactgroup1', 'contactgroup2']
    host_contact_groups = defaultdict(set)

    # c['host_name'] = ['contact1','contact2']
    host_contacts = defaultdict(set)

    # c['host_name'] = '['hostgroup1.hostgroup_name','hostgroup2.hostgroup_name']
    host_hostgroups = defaultdict(set)

    # c['host_name'] = '['service1.get_id()','service2.get_id()']
    host_services = defaultdict(set)

    # c['hostgroup_name'] = ['host_name1','host_name2']
    hostgroup_hosts = defaultdict(set)

    # c['hostgroup_name'] = ['hostgroup1','hostgroup2']
    hostgroup_hostgroups = defaultdict(set)

    # c['hostgroup_name'] = ['service1.get_id()','service2.get_id()']
    hostgroup_services = defaultdict(set)

    # c['service.get_id()'] = '['contactgroup_name1','contactgroup_name2']
    service_contact_groups = defaultdict(set)

    # c['service.get_id()'] = '['contact_name1','contact_name2']
    service_contacts = defaultdict(set)

    # c['service.get_id()'] = '['hostgroup_name1','hostgroup_name2']
    service_hostgroups = defaultdict(set)

    # c['service.get_id()'] = '['servicegroup_name1','servicegroup_name2']
    service_servicegroups = defaultdict(set)

    # c['service.get_id()'] = ['host_name1','host_name2']
    service_hosts = defaultdict(set)

    # c['servicegroup_name'] = ['service1.get_id()', ['service2.get_id()']
    servicegroup_services = defaultdict(set)

    # c[command_name] = '['service.get_id()','service.get_id()']
    command_service = defaultdict(set)

    # c[command_name] = '['host_name1','host_name2']
    command_host = defaultdict(set)

    # use['host']['host_name1'] = ['host_name2','host_name3']
    # use['contact']['contact_name1'] = ['contact_name2','contact_name3']
    _defaultdict_set = lambda: defaultdict(set)
    use = defaultdict( _defaultdict_set )

    # contactgroup_subgroups['contactgroup_name'] = ['group1_name','group2_name']
    contactgroup_subgroups = defaultdict(set)

    @staticmethod
    def reset():
        """ Runs clear() on every member attribute in ObjectRelations """
        for k,v in ObjectRelations.__dict__.items():
            if type(v) == type(defaultdict):
                v.clear()
    @staticmethod
    def _get_subgroups(group_name, dictname):
        """ Helper function that lets you get all sub-group members of a particular group

        For example this call:
          _get_all_group-members('admins', ObjectRelations.contactgroup_contacgroups)

        Will return recursively go through contactgroup_members of 'admins' and return a list
        of  all subgroups
        """
        subgroups = dictname[group_name].copy()
        checked_groups = set()
        while len(subgroups) > 0:
            i = subgroups.pop()
            if i not in checked_groups:
                checked_groups.add(i)
                subgroups.update( dictname[i] )
        return checked_groups

    @staticmethod
    def resolve_contactgroups():
        """ Update all contactgroup relations to take into account contactgroup.contactgroup_members """
        groups = ObjectRelations.contactgroup_contactgroups.keys()
        for group in groups:
            subgroups = ObjectRelations._get_subgroups(group, ObjectRelations.contactgroup_contactgroups)
            ObjectRelations.contactgroup_subgroups[group] = subgroups

            # Loop through every subgroup and apply its attributes to ours
            for subgroup in subgroups:
                for i in ObjectRelations.contactgroup_contacts[subgroup]:
                    ObjectRelations.contact_contactgroups[i].add( group )
                for i in ObjectRelations.contactgroup_hosts[subgroup]:
                    ObjectRelations.host_contact_groups[i].add( group )
                for i in ObjectRelations.contactgroup_services[subgroup]:
                    ObjectRelations.service_contact_groups[i].add( group )
                ObjectRelations.contactgroup_contacts[group].update( ObjectRelations.contactgroup_contacts[subgroup] )
                ObjectRelations.contactgroup_hosts[group].update( ObjectRelations.contactgroup_hosts[subgroup] )
                ObjectRelations.contactgroup_services[group].update( ObjectRelations.contactgroup_services[subgroup] )

    @staticmethod
    def resolve_hostgroups():
        """ Update all hostgroup relations to take into account hostgroup.hostgroup_members """
        groups = ObjectRelations.hostgroup_hostgroups.keys()
        for group in groups:
            subgroups = ObjectRelations._get_subgroups(group, ObjectRelations.hostgroup_hostgroups)
            ObjectRelations.hostgroup_subgroups[group] = subgroups

            # Loop through every subgroup and apply its attributes to ours
            for subgroup in subgroups:
                for i in ObjectRelations.hostgroup_hosts[subgroup]:
                    ObjectRelations.host_hostgroups[i].add( group )
                for i in ObjectRelations.hostgroup_services[subgroup]:
                    ObjectRelations.service_hostgroups[i].add( group )
                ObjectRelations.hostgroup_hosts[group].update( ObjectRelations.hostgroup_hosts[subgroup] )
                ObjectRelations.hostgroup_services[group].update( ObjectRelations.hostgroup_services[subgroup] )


class ObjectFetcher(object):
    """
    This class is a wrapper around pynag.Parsers.config. Is responsible for fetching dict objects
    from config.data and turning into high ObjectDefinition objects

    Internal variables:
     _cached_objects = List of every ObjectDefinition
     _cached_id[o.get_id()] = o
     _cached_shortnames[o.object_type][o.get_shortname()] = o
     _cached_names[o.object_type][o.name] = o
     _cached_object_type[o.object_type].append( o )
    """
    _cached_objects = []
    _cached_ids = {}
    _cached_shortnames = defaultdict(dict)
    _cached_names = defaultdict(dict)
    _cached_object_type = defaultdict(list)

    def __init__(self, object_type):
        self.object_type = object_type

    def get_all(self):
        """ Return all object definitions of specified type"""
        if self.needs_reload():
            self.reload_cache()
        if self.object_type is not None:
            return ObjectFetcher._cached_object_type[self.object_type]
        else:
            return ObjectFetcher._cached_objects

    all = property(get_all)

    def reload_cache(self):
        """Reload configuration cache"""
        # clear object list
        ObjectFetcher._cached_objects = []
        ObjectFetcher._cached_ids = {}
        ObjectFetcher._cached_shortnames = defaultdict(dict)
        ObjectFetcher._cached_names = defaultdict(dict)
        ObjectFetcher._cached_object_type = defaultdict(list)
        global config
        if not config:
            config = Parsers.config(cfg_file)
        if config.needs_reparse():
            debug('Debug: Doing a reparse of configuration')
            config.parse()
        # Reset our list of how objects are related to each other
        ObjectRelations.reset()

        # Fetch all objects from Parsers.config
        for object_type, objects in config.data.items():
            # change "all_host" to just "host"
            object_type = object_type[ len("all_"): ]
            Class = string_to_class.get( object_type, ObjectDefinition )
            for i in objects:
                i = Class(item=i)
                ObjectFetcher._cached_objects.append(i)
                ObjectFetcher._cached_object_type[object_type].append(i)
                ObjectFetcher._cached_ids[i.get_id()] = i
                ObjectFetcher._cached_shortnames[i.object_type][i.get_shortname()] = i
                if i.name is not None:
                    ObjectFetcher._cached_names[i.object_type][i.name] = i
                i._do_relations()
        ObjectRelations.resolve_contactgroups()
        ObjectRelations.resolve_hostgroups()
        return True

    def needs_reload(self):
        """ Returns true if configuration files need to be reloaded/reparsed """
        if ObjectFetcher._cached_objects == []:
            # we get here on first run
            return True
        elif config is None or config.needs_reparse():
            # We get here if any configuration file has changed
            return True
        return False

    def get_by_id(self, id):
        """ Get one specific object

        Returns:
            ObjectDefinition
        Raises:
            ValueError if object is not found
        """
        if self.needs_reload():
            self.reload_cache()
        id = str(id)
        return ObjectFetcher._cached_ids[id]

    def get_by_shortname(self, shortname):
        """ Get one specific object by its shortname (i.e. host_name for host, etc)

        Returns:
            ObjectDefinition
        Raises:
            ValueError if object is not found
        """
        if self.needs_reload():
            self.reload_cache()
        shortname = str(shortname)
        return ObjectFetcher._cached_shortnames[self.object_type][shortname]

    def get_by_name(self, object_name):
        """ Get one specific object by its object_name (i.e. name attribute)

        Returns:
            ObjectDefinition
        Raises:
            ValueError if object is not found
        """
        if self.needs_reload():
            self.reload_cache()
        return ObjectFetcher._cached_names[self.object_type][object_name]

    def get_object_types(self):
        """ Returns a list of all discovered object types """
        if config is None: self.reload_cache()
        return config.get_object_types()

    def filter(self, **kwargs):
        """
        Returns all objects that match the selected filter

        Examples:
        # Get all services where host_name is examplehost.example.com
        Service.objects.filter(host_name='examplehost.example.com')

        # Get service with host_name=examplehost.example.com and service_description='Ping'
        Service.objects.filter(host_name='examplehost.example.com',service_description='Ping')

        # Get all services that are registered but without a host_name
        Service.objects.filter(host_name=None,register='1')

        # Get all hosts that start with 'exampleh'
        Host.objects.filter(host_name__startswith='exampleh')

        # Get all hosts that end with 'example.com'
        Service.objects.filter(host_name__endswith='example.com')

        # Get all contactgroups that contain 'dba'
        Contactgroup.objects.filter(host_name__contains='dba')

        # Get all hosts that are not in the 'testservers' hostgroup
        Host.objects.filter(hostgroup_name__notcontains='testservers')

        # Get all services with non-empty name
        Service.objects.filter(name__isnot=None)

        # Get all hosts that have an address:
        Host.objects.filter(address_exists=True)

        """
        result = []
        # Lets convert all values to str()
        tmp = {}
        for k,v in kwargs.items():
            k = str(k)
            if v is not None: v = str(v)
            tmp[k] = v
        kwargs = tmp
        for i in self.all:
            object_matches = True
            for k, v in kwargs.items():
                if k.endswith('__exists'):
                    k = k[:-len('__exists')]
                    object_matches = str(i.has_key(k)) == str(v)
                    break
                elif k.endswith('__startswith'):
                    k = k[:-12]
                    match_function = str.startswith
                elif k.endswith('__endswith'):
                    k = k[:-10]
                    match_function = str.endswith
                elif k.endswith('__isnot'):
                    k = k[:-7]
                    object_matches = str(i[k]) != str(v)
                    break
                    match_function = str.__ne__
                elif k.endswith('__contains'):
                    k = k[:-10]
                    match_function = contains
                elif k.endswith('__has_field'):
                    k = k[:-11]
                    match_function = has_field
                elif k.endswith('__notcontains'):
                    k = k[:-13]
                    match_function = not_contains
                else:
                    match_function = str.__eq__
                if k == 'id' and str(v) == str(i.get_id()):
                    object_matches = True
                    break
                if k == 'register' and v == '1' and not i.has_key(k):
                    # not defined means item is registered
                    continue
                if v is None and i.has_key(k):
                    object_matches = False
                    break
                if not i.has_key(k):
                    if v is None: continue # if None was the search attribute
                    object_matches = False
                    break
                if not match_function(i[k], v):
                    object_matches = False
                    break
            if object_matches:
                result.append(i)
        return result        

    
class ObjectDefinition(object):
    """
    Holds one instance of one particular Object definition
    Example usage:
        objects = ObjectDefinition.objects.all
        my_object ObjectDefinition( dict ) # dict = hash map of configuration attributes
    """
    object_type = None
    objects = ObjectFetcher(None)

    def __init__(self, item=None, filename=None):
        # Check if we have parsed the configuration yet
        if config is None:
            self.objects.reload_cache()
        # if Item is empty, we are creating a new object
        if item is None:
            item = config.get_new_item(object_type=self.object_type, filename=filename)
            self.is_new = True
        else:
            self.is_new = False
        
        # store the object_type (i.e. host,service,command, etc)
        self.object_type = item['meta']['object_type']
        
        # self.objects is a convenient way to access more objects of the same type
        self.objects = ObjectFetcher(self.object_type)
        # self.data -- This dict stores all effective attributes of this objects
        self._original_attributes = item
        
        #: _changes - This dict contains any changed (but yet unsaved) attributes of this object 
        self._changes = {}
        
        #: _defined_attributes - All attributes that this item has defined
        self._defined_attributes = item['meta']['defined_attributes']
        
        #: _inherited_attributes - All attributes that this object has inherited via 'use'
        self._inherited_attributes = item['meta']['inherited_attributes']
        
        #: _meta - Various metadata about the object
        self._meta = item['meta']
        
        #: _macros - A dict object that resolves any particular Nagios Macro (i.e. $HOSTADDR$)
        self._macros = {}
        
        #: __argument_macros - A dict object that resolves $ARG* macros
        self.__argument_macros = {}

        # Lets find common attributes that every object definition should have:
        self._add_property('register')
        self._add_property('name')
        self._add_property('use')
        
        defs = all_attributes.object_definitions.get(self.object_type,{})
        for k in defs.keys():
            self._add_property(k)
    
    def _add_property(self, name):
        """ Creates dynamic properties for every attribute of out definition.

        i.e. this makes sure host_name attribute is accessable as self.host_name

        Returns: None
        """
        fget = lambda self: self.get_attribute(name)
        fset = lambda self, value: self.set_attribute(name, value)
        fdel = lambda self: self.set_attribute(name, None)
        fdoc = "This is the %s attribute for object definition"
        setattr( self.__class__, name, property(fget,fset,fdel,fdoc))

    def get_attribute(self, attribute_name):
        """Get one attribute from our object definition"""
        return self[attribute_name]

    def set_attribute(self, attribute_name, attribute_value):
        """Set (but does not save) one attribute in our object"""
        self[attribute_name] = attribute_value
        self._event(level="debug", message="attribute changed: %s = %s" % (attribute_name, attribute_value))

    def is_dirty(self):
        """Returns true if any attributes has been changed on this object, and therefore it needs saving"""
        return len(self._changes.keys()) == 0

    def is_registered(self):
        """ Returns true if object is enabled (registered)
        """
        if not self.has_key('register'): return True
        if self['register'] is "1": return True
        return False

    def __setitem__(self, key, item):
        self._changes[key] = item

    def __getitem__(self, key):
        if key == 'id':
            return self.get_id()
        if key == 'description':
            return self.get_description()
        if key == 'shortname':
            return self.get_shortname()
        if key == 'register' and not self._defined_attributes.has_key('register'):
            return "1"
        if key == 'meta':
            return self._meta
        if self._changes.has_key(key):
            return self._changes[key]
        elif self._defined_attributes.has_key(key):
            return self._defined_attributes[key]
        elif self._inherited_attributes.has_key(key):
            return self._inherited_attributes[key]
        elif self._meta.has_key(key):
            return self._meta[key]
        else:
            return None

    def has_key(self, key):
        if key in self.keys():
            return True
        if key in self._meta.keys():
            return True
        return False

    def keys(self):
        all_keys = ['meta']
        for k in self._changes.keys():
            if k not in all_keys: all_keys.append(k)
        for k in self._defined_attributes.keys():
            if k not in all_keys: all_keys.append(k)
        for k in self._inherited_attributes.keys():
            if k not in all_keys: all_keys.append(k)
        #for k in self._meta.keys():
        #    if k not in all_keys: all_keys.append(k)
        return all_keys

    def get_id(self):
        """ Return a unique ID for this object"""
        object_type = self['object_type']
        shortname = self.get_description()
        object_name = self['name']
        filename = self['filename']
        object_id = "%s-%s-%s-%s" % ( object_type, shortname, object_name, filename)
        return str(object_id.__hash__())

    def get_suggested_filename(self):
        """Returns a suitable configuration filename to store this object in

        Typical result value: str('/etc/nagios/pynag/templates/hosts.cfg')
        """
        object_type = self.object_type
        shortname = self.get_shortname()
        # if pynag_directory is undefined, use "/pynag" dir under nagios.cfg
        global pynag_directory
        if pynag_directory is None:
            from os.path import dirname
            pynag_directory = dirname(config.cfg_file) + "/pynag"
        # templates go to the template directory
        if self['register'] == "0":
            path = "%s/templates/%ss.cfg" % (pynag_directory, object_type)
        # Services go to same file as their host
        elif object_type == 'service' and self.host_name is not None:
            try:
                host = self.get_effective_hosts()[0]
                path = host.get_filename()
            except IndexError:
                path = "%s/hosts/%s.cfg" % (pynag_directory, self.host_name)
        else:
            path = "%s/%ss/%s.cfg" % (pynag_directory, object_type, shortname)
        return path

    def save(self):
        """Saves any changes to the current object to its configuration file
        
        Returns:
            Number of changes made to the object
        """
        if config is None: self.objects.reload_cache()
        # Let event-handlers know we are about to save an object
        self._event(level='pre_save', message="%s '%s'." % (self.object_type, self['shortname'] ))
        # If this is a new object, we save it with config.item_add()
        number_of_changes = len(self._changes.keys())
        if self.is_new is True or self.get_filename() is None:
            if not self.get_filename():
                # discover a new filename
                self.set_filename( self.get_suggested_filename() )
            for k,v in self._changes.items():
                self._defined_attributes[k] = v
                self._original_attributes[k] = v
                del self._changes[k]
            config.item_add(self._original_attributes, self.get_filename())
        else:
            # If we get here, we are making modifications to an object
            number_of_changes = 0
            for field_name, new_value in self._changes.items():
                save_result = config.item_edit_field(item=self._original_attributes, field_name=field_name, new_value=new_value)
                if save_result == True:
                    del self._changes[field_name]
                    self._event(level='write', message="%s changed from '%s' to '%s'" % (field_name, self[field_name], new_value))
                    if not new_value:
                        if self._defined_attributes.has_key(field_name):
                            del self._defined_attributes[field_name]
                        if self._original_attributes.has_key(field_name):
                            del self._original_attributes[field_name]
                    else:
                        self._defined_attributes[field_name] = new_value
                        self._original_attributes[field_name] = new_value
                    number_of_changes += 1
                else:
                    raise Exception("Failure saving object. filename=%s, object=%s" % (self.get_filename(), self['shortname']) )

        # this piece of code makes sure that when we current object contains all current info
        self.reload_object()
        self._event(level='save', message="%s '%s' saved." % (self.object_type, self['shortname'] ))
        return number_of_changes
    
    def reload_object(self):
        """ Re applies templates to this object (handy when you have changed the use attribute """
        old_me = config.get_new_item(self.object_type, self.get_filename())
        old_me['meta']['defined_attributes'] = self._defined_attributes
        for k,v in self._defined_attributes.items():
            old_me[k] = v
        for k,v in self._changes.items():
            old_me[k] = v
        i = config._apply_template(old_me)
        new_me = self.__class__(item=i)
        self._defined_attributes = new_me._defined_attributes
        self._original_attributes = new_me._original_attributes
        self._inherited_attributes = new_me._inherited_attributes
        self._meta = new_me._meta
        
    def rewrite(self, str_new_definition=None):
        """ Rewrites this Object Definition in its configuration files.
        
        Arguments:
            str_new_definition = the actual string that will be written in the configuration file
            if str_new_definition is None, then we will use self.__str__()
        Returns: 
            True on success
        """
        self._event(level='pre_save', message="Object definition is being rewritten")
        if str_new_definition is None:
            str_new_definition = self._meta.get('raw_definition')
        config.item_rewrite(self._original_attributes, str_new_definition)
        self['meta']['raw_definition'] = str_new_definition
        self._event(level='write', message="Object definition rewritten")

        # this piece of code makes sure that when we current object contains all current info
        self.reload_object()
        self._event(level='save', message="Object definition was rewritten")
        return True

    def delete(self, recursive=False ):
        """ Deletes this object definition from its configuration files.
        
        Arguments:
            recursive: If True, look for items that depend on this object and delete them as well
            (for example, if you delete a host, delete all its services as well)
        """
        self._event(level="pre_save", message="%s '%s' will be deleted." %( self.object_type, self.get_shortname()))
        if recursive == True:
            # Recursive does not have any meaning for a generic object, this should subclassed.
            pass
        result = config.item_remove(self._original_attributes)
        self._event(level="write", message="%s '%s' was deleted." %( self.object_type, self.get_shortname()))
        self._event(level="save", message="%s '%s' was deleted." %( self.object_type, self.get_shortname()))
        return result

    def copy(self, recursive=False,filename=None, **args):
        """ Copies this object definition with any unsaved changes to a new configuration object
        
        Arguments:
          filename: If specified, new object will be saved in this file.
          recursive: If true, also find any related children objects and copy those
          **args: Any argument will be treated a modified attribute in the new definition.
        Examples:
          myhost = Host.objects.get_by_shortname('myhost.example.com')
          
          # Copy this host to a new one
          myhost.copy( host_name="newhost.example.com", address="127.0.0.1")
          
          # Copy this host and all its services:
                    myhost.copy(recursive=True, host_name="newhost.example.com", address="127.0.0.1")
          
        Returns:
          * A copy of the new ObjectDefinition
          * A list of all copies objects if recursive == True
        """
        if args == {}:
                raise ValueError('To copy an object definition you need at least one new attribute')
        
        new_object = string_to_class[self.object_type]( filename=filename )
        for k,v in self._defined_attributes.items():
            new_object[k] = v
        for k,v in self._changes.items():
            new_object[k] = v
        for k,v in args.items():
            new_object[k] = v
        new_object.save()
        return new_object

        
    def get_related_objects(self):
        """ Returns a list of ObjectDefinition that depend on this object
        
        Object can "depend" on another by a 'use' or 'host_name' or similar attribute
        
        Returns:
            List of ObjectDefinition objects
        """
        result = []
        if self['name'] is not None:
            tmp = ObjectDefinition.objects.filter(use__has_field=self['name'], object_type=self['object_type'])
            for i in tmp: result.append(i)
        return result

    def __str__(self):
        return_buffer = "define %s {\n" % self.object_type
        fields = self._defined_attributes.keys()
        for i in self._changes.keys():
            if i not in fields: fields.append(i)
        fields.sort()
        interesting_fields = ['service_description', 'use', 'name', 'host_name']
        for i in interesting_fields:
            if i in fields:
                fields.remove(i)
                fields.insert(0, i)           
        for key in fields:
            if key == 'meta' or key in self['meta'].keys(): continue
            value = self[key]
            return_buffer += "  %-30s %s\n" % (key, value)
        return_buffer += "}\n"
        return return_buffer

    def __repr__(self):
        return "%s: %s" % (self['object_type'], self.get_shortname())

    def get(self, value, default=None):
        """ self.get(x) == self[x] """
        result = self[value]
        if result is None:
            return default
        else:
            return result

    def get_description(self):
        """ Returns a human friendly string describing current object.

        It will try the following in order:
        * return self get_shortname()
        * return self.name
        * return "Untitled $object_type"
        It defaults to self.get_shortname()
        """
        if self.get_shortname():
           return self.get_shortname()
        if self.name:
            return self.name
        return "Untitled %s" % self.object_type

    def get_shortname(self):
        """ Returns shortname of an object in string format.

        For the confused, nagios documentation refers to shortnames
        usually as <object_type>_name.

        * In case of Host it returns host_name
        * In case of Command it returns command_name
        * etc
        * Special case for services it returns "host_name/service_description"

        Returns None if no attribute can be found to use as a shortname
        """
        return self.get("%s_name" % self.object_type, None)

    def get_filename(self):
        """ Get name of the config file which defines this object
        """
        if self._meta['filename'] is None: return None
        return os.path.normpath( self._meta['filename'] )

    def set_filename(self, filename):
        """ set name of the config file which defines this object"""
        if filename is None:
            self._meta['filename'] = filename
        else:
            self._meta['filename'] = os.path.normpath( filename )

    def get_macro(self, macroname, host_name=None ):
        """ Take macroname (e.g. $USER1$) and return its actual value

        Arguments:
          macroname -- Macro that is to be resolved. For example $HOSTADDRESS$
          host_name -- Optionally specify host (use this for services that
                    -- don't define host specifically for example ones that only
                    -- define hostgroups
        Returns:
          (str) Actual value of the macro. For example "$HOSTADDRESS$" becomes "127.0.0.1"
        """
        if macroname.startswith('$ARG'):
            # Command macros handled in a special function
            return self._get_command_macro(macroname)
        if macroname.startswith('$USER'):
            # $USERx$ macros are supposed to be private, but we will display them anyway
            return config.get_resource(macroname)
        if macroname.startswith('$HOST') or macroname.startswith('$_HOST'):
            return self._get_host_macro(macroname, host_name=host_name)
        if macroname.startswith('$SERVICE') or macroname.startswith('$_SERVICE'):
            return self._get_service_macro(macroname)
        if _standard_macros.has_key( macroname ):
            attr = _standard_macros[ macroname ]
            return self[ attr ]
        return ''

    def get_all_macros(self):
        """Returns {macroname:macrovalue} hash map of this object's macros"""
        # TODO: This function is incomplete and untested
        if self['check_command'] is None: return {}
        c = self['check_command']
        c = c.split('!')
        command_name = c.pop(0)
        command = Command.objects.get_by_shortname(command_name)
        regex = re.compile("(\$\w+\$)")
        macronames = regex.findall( command['command_line'] )
        
        # Add all custom macros to our list:
        for i in self.keys():
            if not i.startswith('_'): continue
            if self.object_type == 'service':
                i = '$_SERVICE%s$' % (i[1:])
            elif self.object_type == 'host':
                i = '$_HOST%s$' % (i[1:])
            macronames.append( i )
        result = {}
        for i in macronames:
            result[i] = self.get_macro(i)
        return result

    def get_effective_command_line(self, host_name=None):
        """Return a string of this objects check_command with all macros (i.e. $HOSTADDR$) resolved"""
        # TODO: This function is incomplete and untested
        if self['check_command'] is None: return None
        c = self['check_command']
        c = c.split('!')
        command_name = c.pop(0)
        try:
            command = Command.objects.get_by_shortname(command_name)
        except ValueError:
            return None
        return self._resolve_macros(command.command_line, host_name=host_name)
    def _resolve_macros(self, string, host_name=None):
        """ Returns string with every $NAGIOSMACRO$ resolved to actual value.

        Arguments:
            string    -- Arbitary string that contains macros
            host_name -- Optionally supply host_name if this service does not define it
        Example:
        >>> i._resolve_macros('$USER1$/check_ping -H $HOSTADDRESS$')
        '/usr/lib64/nagios/plugins/check_ping -H 127.0.0.1'
        """
        if not string:
            return None
        regex = re.compile("(\$\w+\$)")
        get_macro = lambda x: self.get_macro(x.group(), host_name=host_name)
        result = regex.sub(get_macro, string)
        return result

    def run_check_command(self, host_name=None):
        """Run the check_command defined by this service. Returns return_code,stdout,stderr"""
        
        command = self.get_effective_command_line(host_name=host_name)
        if command is None: return None
        proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,stderr=subprocess.PIPE,)
        stdout, stderr = proc.communicate('through stdin to stdout')
        return proc.returncode,stdout,stderr

    def _get_command_macro(self, macroname):
        """Resolve any command argument ($ARG1$) macros from check_command"""
        # TODO: This function is incomplete and untested
        a = {}
        if a == {}:
            c = self['check_command'].split('!')
            c.pop(0) # First item is the command, we dont need it
            for i, v in enumerate( c ):
                tmp = i+1
                if v.startswith('$') and v.endswith('$') and not v.startswith('$ARG'):
                    v = self.get_macro(v)
                a['$ARG%s$' % tmp] = v
        result = a.get(macroname, '')
        # Our $ARGx$ might contain macros on its own, so lets resolve macros in it:
        result = self._resolve_macros(result)
        return result

    def _get_service_macro(self,macroname):
        # TODO: This function is incomplete and untested
        if macroname.startswith('$_SERVICE'):
            # If this is a custom macro
            name = macroname[9:-1]
            return self["_%s" % name]
        if _standard_macros.has_key( macroname ):
            attr = _standard_macros[ macroname ]
            return self[ attr ]
        return ''

    def _get_host_macro(self, macroname, host_name=None):
        # TODO: This function is incomplete and untested
        if macroname.startswith('$_HOST'):
            # if this is a custom macro
            name = macroname[6:-1]
            return self["_%s" % name]
        if _standard_macros.has_key( macroname ):
            attr = _standard_macros[ macroname ]
            return self[ attr ]
        return ''

    def get_effective_children(self, recursive=False):
        """ Get a list of all objects that inherit this object via "use" attribute
        
        Arguments:
            recursive - If true, include grandchildren as well
        Returns:
            A list of ObjectDefinition objects
        """
        if not self.has_key('name'):
            return []
        name = self['name']
        children = self.objects.filter(use__has_field=name)
        if recursive == True:
            grandchildren = []
            for i in children:
                grandchildren += i.get_effective_children(recursive)
            children += grandchildren
        return children

    def get_effective_parents(self, recursive=False):
        """ Get all objects that this one inherits via "use" attribute
        
        Arguments:
            recursive - If true include grandparents in list to be returned
        Returns:
            a list of ObjectDefinition objects
        """
        # TODO: This function is incomplete and untested
        if self['use'] is None:
            return []
        results = []
        use = self['use'].split(',')
        for parent_name in use:
            results.append( self.objects.get_by_name(parent_name) )
        if recursive is True:
            grandparents = []
            for i in results:
                grandparents.append( i.get_effective_parents(recursive=True))
            results += grandparents
        return results

    def get_attribute_tuple(self):
        """ Returns all relevant attributes in the form of:
        
        (attribute_name,defined_value,inherited_value)
        """
        result = []
        for k in self.keys():
            inher = defin = None 
            if self._inherited_attributes.has_key(k):
                inher = self._inherited_attributes[k]
            if self._defined_attributes.has_key(k):
                defin = self._defined_attributes[k]
            result.append((k, defin, inher))
        return result

    def get_parents(self):
        """Returns an ObjectDefinition list of all parents (via use attribute)"""
        result = []
        if not self['use']: return result
        for parent_name in self['use'].split(','):
            search = self.objects.filter(name=parent_name)
            if len(search) < 1: continue
            result.append(search[0])
        return result

    def unregister(self, recursive=True):
        """ Short for self['register'] = 0 ; self.save() """
        self['register'] = 0
        self.save()
        if recursive is True:
            for i in self.get_related_objects():
                i.unregister()

    def attribute_appendfield(self, attribute_name, value):
        """Convenient way to append value to an attribute with a comma seperated value

        Example:
           >>> print myservice
           define service {
                ...
                contactgroups +alladmins,localadmins
            }
           >>> myservice.attribute_addfield(attribute_name="contactgroups", value='webmasters')
           >>> print myservice
           define service {
                ...
                contactgroups +alladmins,localadmins,webmasters
            }
           """
        aList = AttributeList( self[attribute_name] )
        
        if value not in aList.fields:
            aList.fields.append( value )
            self[attribute_name] = str(aList)
        return

    def attribute_removefield(self, attribute_name, value):
        """Convenient way to remove value to an attribute with a comma seperated value

        Example:
           >>> print myservice
           define service {
                ...
                contactgroups +alladmins,localadmins
            }
           >>> myservice.attribute_removefield(attribute_name="contactgroups", value='localadmins')
           >>> print myservice
           define service {
                ...
                contactgroups +alladmins
            }
           """
        aList = AttributeList(self[attribute_name])

        if value in aList.fields:
            aList.fields.remove(value)
            self[attribute_name] = str(aList)
        return

    def attribute_replacefield(self, attribute_name, old_value, new_value):
        """Convenient way to replace field within an attribute with a comma seperated value

        Example:
           >>> print myservice
           define service {
                ...
                contactgroups +alladmins,localadmins
            }
           >>> myservice.attribute_replacefield(attribute_name="contactgroups", old_value='localadmins', new_value=webmasters)
           >>> print myservice
           define service {
                ...
                contactgroups +alladmins,webmasters
            }
           """
        aList = AttributeList(self[attribute_name])
        
        if old_value in aList.fields:
            i = aList.fields.index(old_value)
            aList.fields[i] = new_value
            self[attribute_name] = str(aList)
        return

    def _get_effective_attribute(self, attribute_name):
        """This helper function returns specific attribute, from this object or its templates
        
        This is handy for fields that effectively are many_to_many values.
        for example, "contactroups +group1,group2,group3"
        
        Fields that are known to use this format are:
            contacts, contactgroups, hostgroups, servicegroups, members,contactgroup_members
        """
        result = []
        tmp = self[attribute_name]
        if tmp is not None:
            result.append( tmp )
        if tmp is None or tmp.startswith('+'):
            for parent in self.get_parents():
                result.append( parent._get_effective_attribute(attribute_name) )
                if parent[attribute_name] is not None and not parent[attribute_name].startswith('+'):
                    break
        return_value = []
        for value in  result :
            value = value.strip('+')
            if value == '': continue
            if value not in return_value:
                return_value.append( value )
        tmp = ','.join( return_value )
        tmp = tmp.replace(',,',',')
        return tmp

    def _event(self, level=None, message=None):
        """ Pass informational message about something that has happened within the Model """
        for i in eventhandlers:
            if level == 'write':
                i.write( object_definition=self, message=message )
            elif level == 'save':
                i.save( object_definition=self, message=message )
            elif level == 'pre_save':
                i.pre_save( object_definition=self, message=message )
            else:
                i.debug( object_definition=self, message=message )

    def _do_relations(self):
        """ Discover all related objects (f.e. services that belong to this host, etc

        ObjectDefinition only has relations via 'use' paramameter. Subclasses should extend this.
        """
        parents = AttributeList( self.use )
        for i in parents.fields:
            ObjectRelations.use[self.object_type][ i ].add( self.get_id() )


class Host(ObjectDefinition):
    object_type = 'host'
    objects = ObjectFetcher('host')

    def get_effective_services(self):
        """ Returns a list of all Service that belong to this Host """
        get_object = lambda x: Service.objects.get_by_id(x)
        list_of_shortnames = ObjectRelations.host_services[self.host_name]
        return map( get_object, list_of_shortnames )

    def get_effective_contacts(self):
        """ Returns a list of all Contact that belong to this Host """
        get_object = lambda x: Contact.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.host_contacts[self.host_name]
        return map( get_object, list_of_shortnames )

    def get_effective_contact_groups(self):
        """ Returns a list of all Contactgroup that belong to this Host """
        get_object = lambda x: Contactgroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.host_contact_groups[self.host_name]
        return map( get_object, list_of_shortnames )

    def get_effective_hostgroups(self):
        """ Returns a list of all Hostgroup that belong to this Host """
        get_object = lambda x: Hostgroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.host_hostgroups[self.host_name]
        return map( get_object, list_of_shortnames )

    def delete(self, recursive=False ):
        """ Overwrites objectdefinition so that recursive=True will delete all services as well """
        # Find all services and delete them as well
        if recursive == True:
            for i in self.get_effective_services():
                i.delete(recursive=recursive)
        # Call parent to get delete myself
        super(self.__class__, self).delete(recursive=recursive)

    def get_related_objects(self):
        result = super(self.__class__, self).get_related_objects()
        if self['host_name'] is not None:
            tmp = Service.objects.filter(host_name=self['host_name'])
            for i in tmp: result.append( i )
        return result

    def get_effective_check_command(self):
        """ Returns a Command object as defined by check_command attribute

        Raises KeyError if check_command is not found or not defined.
        """
        c = self.check_command
        if not c or c == '':
            raise KeyError(None)
        check_command = c.split('!')[0]
        return Command.objects.get_by_shortname(check_command)
    def get_current_status(self):
        """ Same as Service.get_current_status() """
        return Status(self.host_name)
    def copy(self, recursive=False,filename=None, **args):
        """ Same as ObjectDefinition.copy() except can recursively copy services """
        copies = []
        copies.append( ObjectDefinition.copy(self, recursive=recursive,filename=filename, **args) )
        if recursive == True and 'host_name' in args:
            for i in self.get_effective_services():
                print i.get_shortname()
                copies.append( i.copy(filename=filename, host_name=args.get('host_name')) )
        return copies

    def _do_relations(self):
        super(self.__class__, self)._do_relations()
        # Do hostgroups
        hg = AttributeList( self.hostgroups )
        for i in hg.fields:
            ObjectRelations.host_hostgroups[self.host_name].add( i )
            ObjectRelations.hostgroup_hosts[i].add( self.host_name )
        # Contactgroups
        cg = AttributeList( self.contact_groups )
        for i in cg.fields:
            ObjectRelations.host_contact_groups[self.host_name].add( i )
            ObjectRelations.contactgroup_hosts[i].add( self.get_id() )
        contacts = AttributeList( self.contacts )
        for i in contacts.fields:
            ObjectRelations.host_contacts[self.host_name].add( i )
            ObjectRelations.contact_hosts[i].add( self.get_id() )
        if self.check_command:
            command_name = self.check_command.split('!')[0]
            ObjectRelations.command_service[ self.host_name ].add( command_name )


class Service(ObjectDefinition):
    object_type = 'service'
    objects = ObjectFetcher('service')

    def get_shortname(self):
        if not self.host_name and not self.service_description:
            return None
        return "%s/%s" % (self['host_name'], self['service_description'])

    def _get_host_macro(self, macroname, host_name=None):
        if not host_name:
            host_name = self['host_name']
        if not host_name:
            return None
        try:
            myhost = Host.objects.get_by_shortname(host_name)
            return myhost._get_host_macro(macroname)     
        except Exception:
            return None

    def _do_relations(self):
        super(self.__class__, self)._do_relations()
        # Do hostgroups
        hg = AttributeList( self.hostgroup_name )
        for i in hg.fields:
            ObjectRelations.service_hostgroups[self.get_id()].add( i )
            ObjectRelations.hostgroup_services[i].add( self.get_id() )
            # Contactgroups
        cg = AttributeList( self.contact_groups )
        for i in cg.fields:
            ObjectRelations.service_contact_groups[self.get_id()].add( i )
            ObjectRelations.contactgroup_services[i].add( self.get_id() )
        contacts = AttributeList( self.contacts )
        for i in contacts.fields:
            ObjectRelations.service_contacts[self.get_id()].add( i )
            ObjectRelations.contact_services[i].add( self.get_id() )
        sg = AttributeList( self.servicegroups )
        for i in sg.fields:
            ObjectRelations.service_servicegroups[self.get_id()].add( i )
            ObjectRelations.servicegroup_services[i].add( self.get_id() )
        if self.check_command:
            command_name = self.check_command.split('!')[0]
            ObjectRelations.command_service[ self.get_id() ].add( command_name )
        hosts = AttributeList( self.host_name )
        for i in hosts.fields:
            ObjectRelations.service_hosts[ self.get_id() ].add( i )
            ObjectRelations.host_services[ i ].add( self.get_id() )

    def get_effective_hosts(self):
        """ Returns a list of all Host that belong to this Service """
        get_object = lambda x: Host.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.service_hosts[ self.get_id() ]
        return map( get_object, list_of_shortnames )

    def get_effective_contacts(self):
        """ Returns a list of all Contact that belong to this Service """
        get_object = lambda x: Contact.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.service_contacts[self.get_id()]
        return map( get_object, list_of_shortnames )

    def get_effective_contact_groups(self):
        """ Returns a list of all Contactgroup that belong to this Service """
        get_object = lambda x: Contactgroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.service_contact_groups[self.get_id()]
        return map( get_object, list_of_shortnames )

    def get_effective_hostgroups(self):
        """ Returns a list of all Hostgroup that belong to this Service """
        get_object = lambda x: Hostgroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.service_hostgroups[self.get_id()]
        return map( get_object, list_of_shortnames )

    def get_effective_servicegroups(self):
        """ Returns a list of all Servicegroup that belong to this Service """
        get_object = lambda x: Servicegroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.service_servicegroups[self.get_id()]
        return map( get_object, list_of_shortnames )

    def get_effective_check_command(self):
        """ Returns a Command object as defined by check_command attribute

        Raises KeyError if check_command is not found or not defined.
        """
        c = self.check_command
        if not c or c == '':
            raise KeyError(None)
        check_command = c.split('!')[0]
        return Command.objects.get_by_shortname(check_command)
    def get_current_status(self):
        """ Returns a Status object, reflecting this object status (i.e. status.dat)
        """
        return Status(self.host_name, self.service_description)
class Command(ObjectDefinition):
    object_type = 'command'
    objects = ObjectFetcher('command')


class Contact(ObjectDefinition):
    object_type = 'contact'
    objects = ObjectFetcher('contact')

    def get_description(self):
        """ Returns a friendly description of the object """
        return self['contact_name']

    def get_effective_contactgroups(self):
        """ Get a list of all Contactgroup that are hooked to this contact """
        get_object = lambda x: Contactgroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.contact_contactgroups[self.contact_name]
        return map( get_object, list_of_shortnames )

    def get_effective_hosts(self):
        """ Get a list of all Host that are hooked to this Contact """
        result = set()
        # First add all hosts that name this contact specifically
        get_object = lambda x: Host.objects.get_by_id(x)
        list_of_shortnames = ObjectRelations.contact_hosts[self.contact_name]
        result.update( map( get_object, list_of_shortnames ) )

        # Next do the same for all contactgroups this contact belongs in
        for i in self.get_effective_contactgroups():
            result.update( i.get_effective_hosts() )
        return result

    def get_effective_services(self):
        """ Get a list of all Service that are hooked to this Contact """
        result = set()
        # First add all services that name this contact specifically
        get_object = lambda x: Service.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.contact_services[self.contact_name]
        result.update( map( get_object, list_of_shortnames ) )
        return result

    def _do_relations(self):
        super(self.__class__, self)._do_relations()
        groups = AttributeList( self.contactgroups )
        for i in groups.fields:
            ObjectRelations.contact_contactgroups[self.contact_name].add( i )
            ObjectRelations.contactgroup_contacts[i].add( self.contact_name )


class ServiceDependency(ObjectDefinition):
    object_type = 'servicedependency'
    objects = ObjectFetcher('servicedependency')


class HostDependency(ObjectDefinition):
    object_type = 'hostdependency'
    objects = ObjectFetcher('hostdependency')


class Contactgroup(ObjectDefinition):
    object_type = 'contactgroup'
    objects = ObjectFetcher('contactgroup')

    def get_effective_contactgroups(self):
        """ Returns a list of every Contactgroup that is a member of this Contactgroup """
        get_object = lambda x: Contactgroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.contactgroup_subgroups[self.contactgroup_name]
        return map( get_object, list_of_shortnames )

    def get_effective_contacts(self):
        """ Returns a list of every Contact that is a member of this Contactgroup """
        get_object = lambda x: Contact.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.contactgroup_contacts[self.contactgroup_name]
        return map( get_object, list_of_shortnames )

    def get_effective_hosts(self):
        """ Return every Host that belongs to this contactgroup """
        list_of_shortnames = ObjectRelations.contactgroup_hosts[self.contactgroup_name]
        get_object = lambda x: Host.objects.get_by_id(x)
        return map( get_object, list_of_shortnames )

    def get_effective_services(self):
        """ Return every Host that belongs to this contactgroup """
        services = {}
        for i in Service.objects.all:
            services[i.get_id()] = i
        list_of_shortnames = ObjectRelations.contactgroup_services[self.contactgroup_name]
        get_object = lambda x: services[x]
        return map( get_object, list_of_shortnames )

    def _do_relations(self):
        super(self.__class__, self)._do_relations()
        members = AttributeList( self.members )
        for i in members.fields:
            ObjectRelations.contactgroup_contacts[self.contactgroup_name].add( i )
            ObjectRelations.contact_contactgroups[i].add( self.contactgroup_name )
        groups = AttributeList( self.contactgroup_members )
        for i in groups.fields:
            ObjectRelations.contactgroup_contactgroups[self.contactgroup_name].add( i )

   
class Hostgroup(ObjectDefinition):
    object_type = 'hostgroup'
    objects = ObjectFetcher('hostgroup')

    def get_effective_services(self):
        """ Returns a list of all Service that belong to this hostgroup """
        list_of_shortnames = ObjectRelations.hostgroup_services[self.hostgroup_name]
        get_object = lambda x: Service.objects.get_by_id(x)
        return map( get_object, list_of_shortnames )

    def get_effective_hosts(self):
        """ Returns a list of all Host that belong to this hostgroup """
        list_of_shortnames = ObjectRelations.hostgroup_hosts[self.hostgroup_name]
        get_object = lambda x: Host.objects.get_by_shortname(x)
        return map( get_object, list_of_shortnames )

    def get_effective_hostgroups(self):
        """ Returns a list of every Hostgroup that is a member of this Hostgroup """
        get_object = lambda x: Hostgroup.objects.get_by_shortname(x)
        list_of_shortnames = ObjectRelations.hostgroup_subgroups[self.hostgroup_name]
        return map( get_object, list_of_shortnames )

    def _do_relations(self):
        super(self.__class__, self)._do_relations()
        members = AttributeList( self.members )
        for i in members.fields:
            ObjectRelations.hostgroup_hosts[self.hostgroup_name].add( i )
            ObjectRelations.host_hostgroups[i].add( self.hostgroup_name )
        groups = AttributeList( self.hostgroup_members )
        for i in groups.fields:
            ObjectRelations.hostgroup_hostgroups[self.hostgroup_name].add( i )


class Servicegroup(ObjectDefinition):
    object_type = 'servicegroup'
    objects = ObjectFetcher('servicegroup')

    def get_effective_services(self):
        """ Returns a list of all Service that belong to this Servicegroup """
        list_of_shortnames = ObjectRelations.servicegroup_services[self.servicegroup_name]
        get_object = lambda x: Service.objects.get_by_id(x)
        return map( get_object, list_of_shortnames )


class Timeperiod(ObjectDefinition):
    object_type = 'timeperiod'
    objects = ObjectFetcher('timeperiod')

class Status(object):
    """ Contains current status info (i.e. status.dat) for one specific Host or Service """
    plugin_output = ''
    perfdata = ''
    long_plugin_output = ''
    current_state = ''
    perfdatalist = []
    def __init__(self, host_name, service_description=None):
        self.host_name = host_name
        s = Parsers.status()
        s.parse()
        try:
            if not service_description:
                # This is a host status
                self.status = s.get_hoststatus(host_name)
            else:
                # This is a service status
                self.status = s.get_servicestatus(host_name, service_description)
        except ValueError:
            self.status = None
        self.plugin_output = self.status['plugin_output']
        self.perfdata = self.status['performance_data']
        self.long_plugin_output = self.status['long_plugin_output']
        self.current_state = self.status['current_state']
        self.perfdatalist = PerfData(self.perfdata)

class AttributeList(object):
    """ Parse a list of nagios attributes (e. contact_groups) into a parsable format

    This makes it handy to mangle with nagios attribute values that are in a comma seperated format.

    Typical comma-seperated format in nagios configuration files looks something like this:
        contact_groups     +group1,group2,group3

    Example:
        >>> i = AttributeList('+group1,group2,group3')
        >>> print "Operator is:", i.operator
        Operator is: +
        >>> print i.values
        ['group1','group2','group3']
    """

    def __init__(self, value=None):
        self.operator = ''
        self.fields = []
        
        # this is easy to do if attribue_name is unset
        if not value or value == '':
            return
    
        possible_operators = '+-!'
        if value[0] in possible_operators:
            self.operator = value[0]
        else:
            self.operator = ''
        
        self.fields = value.strip(possible_operators).split(',')

    def __str__(self):
        return self.operator + ','.join(self.fields)

    def __repr__(self):
        return self.__str__()

    def insert(self, index, object):
        return self.fields.insert(index,object) 

    def append(self, object):
        return self.fields.append(object)

    def count(self, value):
        return self.fields.count(value)

    def extend(self, iterable):
        return self.fields.extend(iterable)

    def index(self, value, start, stop):
        return self.fields.index(value, start, stop)

    def reverse(self):
        return self.fields.reverse()

    def sort(self):
        return self.fields.sort()

    def remove(self, value):
        return self.fields.remove(value)

string_to_class = {}
string_to_class['contact'] = Contact
string_to_class['service'] = Service
string_to_class['host'] = Host
string_to_class['hostgroup'] = Hostgroup
string_to_class['contactgroup'] = Contactgroup
string_to_class['servicegroup'] = Servicegroup
string_to_class['timeperiod'] = Timeperiod
string_to_class['hostdependency'] = HostDependency
string_to_class['servicedependency'] = ServiceDependency
string_to_class['command'] = Command
#string_to_class[None] = ObjectDefinition


if __name__ == '__main__':
    o = Service.objects.all[17]
    print o['__EXTRAOPTS']
    print o.get_effective_command_line()
    o['__EXTRAOPTS'] = "new value"
    print o.get_effective_command_line()



# -*- coding: utf-8 -*-
#
# pynag - Python Nagios plug-in and configuration environment
# Copyright (C) 2010 Drew Stinnet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import re
import time


def debug(text):
    debug = True
    if debug: print text


class config:
    """
    Parse and write nagios config files
    """
    def __init__(self, cfg_file=None):

        self.cfg_file = cfg_file  # Main configuration file

        # If nagios.cfg is not set, lets do some minor autodiscover.
        if self.cfg_file is None:
            self.cfg_file = self.guess_cfg_file()
        self.data = {}
    def guess_cfg_file(self):
        """ Returns a path to any nagios.cfg found on your system

        Use this function if you don't want specify path to nagios.cfg in your
        code and you are confident that it is located in a common location
        """
        possible_files = ('/etc/nagios/nagios.cfg',
                          '/etc/nagios3/nagios.cfg',
                          '/usr/local/nagios/etc/nagios.cfg',
                          '/nagios/etc/nagios/nagios.cfg',
                          './nagios.cfg',
                          './nagios/nagios.cfg',
            )
        for file in possible_files:
            if os.path.isfile(file):
                return file
        raise ParserError('Could not find nagios.cfg')
    def reset(self):
        self.cfg_files = [] # List of other configuration files
        self.data = {} # dict of every known object definition
        self.errors = [] # List of ParserErrors
        self.item_list = None
        self.item_cache = None
        self.maincfg_values = [] # The contents of main nagios.cfg
        self._resource_values = [] # The contents of any resource_files

        ## This is a pure listof all the key/values in the config files.  It
        ## shouldn't be useful until the items in it are parsed through with the proper
        ## 'use' relationships
        self.pre_object_list = []
        self.post_object_list = []
        self.object_type_keys = {
            'hostgroup':'hostgroup_name',
            'hostextinfo':'host_name',
            'host':'host_name',
            'service':'name',
            'servicegroup':'servicegroup_name',
            'contact':'contact_name',
            'contactgroup':'contactgroup_name',
            'timeperiod':'timeperiod_name',
            'command':'command_name',
            #'service':['host_name','description'],
        }

    def _has_template(self, target):
        """
        Determine if an item has a template associated with it
        """
        return 'use' in target

    def _get_hostgroup(self, hostgroup_name):
        return self.data['all_hostgroup'].get(hostgroup_name, None)

    def _get_key(self, object_type, user_key = None):
        """
        Return the correct 'key' for an item.  This is mainly a helper method
        for other methods in this class.  It is used to shorten code repitition
        """
        if not user_key and not self.object_type_keys.has_key(object_type):
            raise ParserError("Unknown key for object type:  %s\n" % object_type)

        ## Use a default key
        if not user_key:
            user_key = self.object_type_keys[object_type]

        return user_key

    def _get_item(self, item_name, item_type):
        """ 
           Return an item from a list
           """
        # create local cache for performance optimizations. TODO: Rewrite functions that call this function
        if not self.item_list:
            self.item_list = self.pre_object_list
            self.item_cache = {}
            for item in self.item_list:
                if not item.has_key('name'):
                    continue
                name = item['name']
                tmp_item_type = (item['meta']['object_type'])
                if not self.item_cache.has_key(tmp_item_type):
                    self.item_cache[tmp_item_type] = {}
                self.item_cache[tmp_item_type][name] = item
        return self.item_cache[item_type].get(item_name, None)

    def _apply_template(self, original_item):
        """
        Apply all attributes of item named parent_name to "original_item".
        """
        # If item does not inherit from anyone else, lets just return item as is.
        if not original_item.has_key('use'):
            return original_item
        object_type = original_item['meta']['object_type']
        # Performance tweak, if item has been parsed. Lets not do it again
        if original_item.has_key('name') and self.item_apply_cache[object_type].has_key(original_item['name']):
            return self.item_apply_cache[object_type][ original_item['name'] ]
        # End of performance tweak
        parent_names = original_item['use'].split(',')
        parent_items = []
        for parent_name in parent_names:
            parent_item = self._get_item( parent_name, object_type )
            if parent_item is None:
                error_string = "Can not find any %s named %s\n" % (object_type,parent_name)
                self.errors.append( ParserError(error_string,item=original_item) )
                continue
            # Parent item probably has use flags on its own. So lets apply to parent first
            try:
                parent_item = self._apply_template( parent_item )
            except RuntimeError, e:
                self.errors.append( ParserError("Error while parsing item: %s (it might have circular use=)" % str(e), item=original_item) )
            parent_items.append( parent_item )
        for parent_item in parent_items:
            for k,v in parent_item.iteritems():
                if k == 'use':
                    continue
                if k == 'register':
                    continue
                if k == 'meta':
                    continue
                if k == 'name':
                    continue
                if not original_item['meta']['inherited_attributes'].has_key(k):
                    original_item['meta']['inherited_attributes'][k] = v
                if not original_item.has_key(k):
                    original_item[k] = v
                    original_item['meta']['template_fields'].append(k)
        if original_item.has_key('name'):
            self.item_apply_cache[object_type][ original_item['name'] ] = original_item
        return original_item

    def _get_items_in_file(self, filename):
        """
        Return all items in the given file
        """
        return_list = []

        for k in self.data.keys():
            for item in self[k]:
                if item['meta']['filename'] == filename:
                    return_list.append(item)
        return return_list

    def get_new_item(self, object_type, filename):
        """ Returns an empty item with all necessary metadata """
        current = {}
        current['meta'] = {}
        current['meta']['object_type'] = object_type
        current['meta']['filename'] = filename
        current['meta']['template_fields'] = []
        current['meta']['needs_commit'] = None
        current['meta']['delete_me'] = None
        current['meta']['defined_attributes'] = {}
        current['meta']['inherited_attributes'] = {}
        current['meta']['raw_definition'] = "define %s {\n\n}" % object_type
        return current

    def _load_file(self, filename):
        ## Set globals (This is stolen from the perl module)
        append = ""
        current = None
        in_definition = {}
        tmp_buffer = []

        for line_num, line in enumerate( open(filename, 'rb').readlines() ):

            ## Cleanup and line skips
            line = line.strip()
            if line == "":
                continue
            if line[0] == "#" or line[0] == ';':
                continue

            # TODO: Find out why this code append lives here, are there really any cases
            # Where a nagios attributes expands more than one line ? 
            # append saved text to the current line
            if append:
                append += ' '
                line = append + line
                append = None

            # end of object definition
            if line.find("}") != -1:

                in_definition = None
                current['meta']['line_end'] = line_num
                # Looks to me like nagios ignores everything after the } so why shouldn't we ?
                rest = line.split("}", 1)[1]
                
                tmp_buffer.append(  line )
                try:
                    current['meta']['raw_definition'] = '\n'.join( tmp_buffer )
                except Exception:
                    raise ParserError("Encountered Unexpected end of object definition in file '%s'." % filename)
                self.pre_object_list.append(current)

                ## Destroy the Nagios Object
                current = None
                continue

            # beginning of object definition
            boo_re = re.compile("define\s+(\w+)\s*\{?(.*)$")
            m = boo_re.search(line)
            if m:
                tmp_buffer = [line]
                object_type = m.groups()[0]
                current = self.get_new_item(object_type, filename)
                current['meta']['line_start'] = line_num

                if in_definition:
                    raise ParserError("Error: Unexpected start of object definition in file '%s' on line $line_no.  Make sure you close preceding objects before starting a new one.\n" % filename)

                ## Start off an object
                in_definition = True
                
                # Looks to me like nagios ignores everything after the {, so why shouldn't we ?
                rest = m.groups()[1]
                continue
            else:
                tmp_buffer.append( '    ' + line )

            ## save whatever's left in the buffer for the next iteration
            if not in_definition:
                append = line
                continue

            ## this is an attribute inside an object definition
            if in_definition:
                #(key, value) = line.split(None, 1)
                tmp = line.split(None, 1)
                if len(tmp) > 1:
                    (key, value) = tmp
                else:
                    key = tmp[0]
                    value = ""

                ## Strip out in-line comments
                if value.find(";") != -1:
                    value = value.split(";", 1)[0]

                ## Clean info
                key = key.strip()
                value = value.strip()

                ## Rename some old values that may be in the configuration
                ## This can probably be removed in the future to increase performance
                if (current['meta']['object_type'] == 'service') and key == 'description':
                    key = 'service_description'

                # Special hack for timeperiods as they are not consistent with other objects
                # We will treat whole line as a key with an empty value
                if (current['meta']['object_type'] == 'timeperiod' ) and key not in  ('timeperiod_name', 'alias'):
                    key = line
                    value = ''
                current[key] = value
                current['meta']['defined_attributes'][key] = value
            ## Something is wrong in the config
            else:
                raise ParserError("Error: Unexpected token in file '%s'" % filename)

        ## Something is wrong in the config
        if in_definition:
            raise ParserError("Error: Unexpected EOF in file '%s'" % filename)

    def _locate_item(self, item):
        """
        This is a helper function for anyone who wishes to modify objects. It takes "item", locates the
        file which is configured in, and locates exactly the lines which contain that definition.
        
        Returns tuple:
            (everything_before, object_definition, everything_after, filename)
            everything_before(string) - Every line in filename before object was defined
            everything_after(string) - Every line in "filename" after object was defined
            object_definition - exact configuration of the object as it appears in "filename"
            filename - file in which the object was written to
        Raises:
            ValueError if object was not found in "filename"
        """
        if item['meta'].has_key("filename"):
            filename = item['meta']['filename']
        else:
            raise ValueError("item does not have a filename")
        file = open(filename)
        object_has_been_found = False
        everything_before = [] # Every line before our object definition
        everything_after = []  # Every line after our object definition
        object_definition = [] # List of every line of our object definition
        tmp_buffer = []        # Every line of current object being parsed is stored here.
        current_object_type = None # Object type of current object goes in here
        i_am_within_definition = False
        for line in file.readlines():
            if object_has_been_found:
                # If we have found an object, lets just spool to the end
                everything_after.append( line )
                continue
            tmp = line.split(None, 1)
            if len(tmp) == 0:
                # empty line
                keyword = ''
                rest = ''
            elif len(tmp) == 1:
                # single word on the line
                keyword = tmp[0]
                rest = ''
            else:
                keyword,rest = tmp[0],tmp[1]
            keyword = keyword.strip()
            # If we reach a define statement, we log every line to a special buffer
            # When define closes, we parse the object and see if it is the object we
            # want to modify
            if keyword == 'define':
                current_object_type = rest.split(None,1)[0]
                current_object_type = current_object_type.strip(';')
                current_object_type = current_object_type.strip('{')
                current_object_type = current_object_type.strip()
                tmp_buffer = []
                i_am_within_definition = True
            if i_am_within_definition == True:
                tmp_buffer.append( line )
            else:
                everything_before.append( line )
            if len(keyword) > 0 and keyword[0] == '}':
                i_am_within_definition = False
                
                current_candidate = self.get_new_item(object_type=current_object_type, filename=filename)
                for i in tmp_buffer:
                    i = i.strip()
                    tmp = i.split(None, 1)
                    # Hack that makes timeperiod attributes be contained only in the key
                    if len(tmp) == 0:
                        continue
                    if current_object_type == 'timeperiod' and tmp[0] not in ('alias', 'timeperiod_name'):
                        k = i
                        v = ''
                    elif len(tmp) == 1:
                        k = tmp[0]
                        v = ''
                    elif len(tmp) > 1:
                        k,v = tmp[0],tmp[1]
                        v = v.split(';',1)[0]
                        v = v.strip()
                    else: continue # skip empty lines
                    
                    if k.startswith('#'): continue
                    if k.startswith(';'): continue
                    if k.startswith('define'): continue
                    if k.startswith('}'): continue
                    
                    current_candidate[k] = v
                    current_candidate['meta']['defined_attributes'][k] = v
                    # Apply template should not be needed anymore
                    #current_candidate = self._apply_template(current_candidate)
                # Compare objects
                if self.compareObjects( item, current_candidate ) == True:
                    # This is the object i am looking for
                    object_has_been_found = True
                    object_definition = tmp_buffer
                else:
                    # This is not the item you are looking for
                    everything_before += tmp_buffer
        if object_has_been_found:
            return everything_before, object_definition, everything_after, filename
        else:
            raise ValueError("We could not find object in %s\n%s" % (filename,item))

    def _modify_object(self, item, field_name=None, new_value=None, new_field_name=None, new_item=None, make_comments=True):
        """
        Helper function for object_* functions. Locates "item" and changes the line which contains field_name.
        If new_value and new_field_name are both None, the attribute is removed.

        Arguments:
            item(dict) -- The item to be modified
            field_name(str) -- The field_name to modify (if any)
            new_field_name(str) -- If set, field_name will be renamed
            new_value(str) -- If set the value of field_name will be changed
            new_item(str) -- If set, whole object will be replaced with this string
            make_comments -- If set, put pynag-branded comments where changes have been made
        Returns:
            True on success
        Raises:
            ValueError if object or field_name is not found
            IOError is save is unsuccessful.
        """
        if field_name is None and new_item is None:
            raise ValueError("either field_name or new_item must be set")
        everything_before,object_definition, everything_after, filename = self._locate_item(item)
        if new_item is not None:
            # We have instruction on how to write new object, so we dont need to parse it
            object_definition = [new_item]
        else:
            change = None
            i = 0
            for i in range( len(object_definition)):
                tmp = object_definition[i].split(None, 1)
                if len(tmp) == 0: continue
                # Hack for timeperiods, they dont work like other objects
                elif item['meta']['object_type'] == 'timeperiod' and tmp[0] not in ('alias', 'timeperiod_name'):
                    tmp = [object_definition[i]]
                    # we can't change timeperiod, so we fake a field rename
                    if new_value is not None:
                        new_field_name = new_value
                        new_value = None
                        value = ''
                elif len(tmp) == 1: value = ''
                else: value = tmp[1]
                k = tmp[0].strip()
                if k == field_name:
                    # Attribute was found, lets change this line
                    if new_field_name is None and new_value is None:
                        # We take it that we are supposed to remove this attribute
                        change = object_definition.pop(i)
                        break
                    elif new_field_name:
                        # Field name has changed
                        k = new_field_name
                    if new_value is not None:
                        # value has changed
                        value = new_value
                    # Here we do the actual change    
                    change = "\t%-30s%s\n" % (k, value)
                    if item['meta']['object_type'] == 'timeperiod' and field_name not in ('alias', 'timeperiod_name'):
                        change = "\t%s\n" % (new_field_name)
                    object_definition[i] = change
                    break
            if not change:
                    # Attribute was not found. Lets add it
                    change = "\t%-30s%s\n" % (field_name, new_value)
                    object_definition.insert(i,change)
        # Lets put a banner in front of our item
        if make_comments:
            comment = '# Edited by PyNag on %s\n' % time.ctime()
            if len(everything_before) > 0:
                last_line_before = everything_before[-1]
                if last_line_before.startswith('# Edited by PyNag on'):
                    everything_before.pop() # remove this line
            object_definition.insert(0, comment)
        # Here we overwrite the config-file, hoping not to ruin anything
        buffer = "%s%s%s" % (''.join(everything_before), ''.join(object_definition), ''.join(everything_after))
        file = open(filename,'w')
        file.write( buffer )
        file.close()
        return True        

    def item_rewrite(self, item, str_new_item):
        """
        Completely rewrites item with string provided.
        
        Arguments:
            item -- Item that is to be rewritten
            str_new_item -- str representation of the new item
        Examples:
            item_rewrite( item, "define service {\n name example-service \n register 0 \n }\n" )
        Returns:
            True on success
        Raises:
            ValueError if object is not found
            IOError if save fails
        """
        return self._modify_object(item=item, new_item=str_new_item)

    def item_remove(self, item):
        """
        Completely rewrites item with string provided.
        
        Arguments:
            item -- Item that is to be rewritten
            str_new_item -- str representation of the new item
        Examples:
            item_rewrite( item, "define service {\n name example-service \n register 0 \n }\n" )
        Returns:
            True on success
        Raises:
            ValueError if object is not found
            IOError if save fails
        """
        return self._modify_object(item=item, new_item="")

    def item_edit_field(self, item, field_name, new_value):
        """
        Modifies one field of a (currently existing) object. Changes are immediate (i.e. there is no commit)
        
        Example usage:
            edit_object( item, field_name="host_name", new_value="examplehost.example.com")
        Returns:
            True on success
        Raises:
            ValueError if object is not found
            IOError if save fails
        """
        return self._modify_object(item, field_name=field_name, new_value=new_value)
    
    def item_remove_field(self, item, field_name):
        """
        Removes one field of a (currently existing) object. Changes are immediate (i.e. there is no commit)
        
        Example usage:
            item_remove_field( item, field_name="contactgroups" )
        Returns:
            True on success
        Raises:
            ValueError if object is not found
            IOError if save fails
        """
        return self._modify_object(item=item, field_name=field_name, new_value=None, new_field_name=None)
    
    def item_rename_field(self, item, old_field_name, new_field_name):
        """
        Renames a field of a (currently existing) item. Changes are immediate (i.e. there is no commit).
        
        Example usage:
            item_rename_field(item, old_field_name="normal_check_interval", new_field_name="check_interval")
        Returns:
            True on success
        Raises:
            ValueError if object is not found
            IOError if save fails
        """
        return self._modify_object(item=item, field_name=old_field_name, new_field_name=new_field_name)

    def item_add(self, item, filename):
        """
        Adds a new object to a specified config file
        
        Arguments:
            item -- Item to be created
            filename -- Filename that we are supposed to write to
        Returns:
            True on success
        Raises:
            IOError on failed save
        """
        if not 'meta' in item:
            item['meta'] = {}
        item['meta']['filename'] = filename
        
        # Create directory if it does not already exist                
        dirname = os.path.dirname(filename)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        buffer = self.print_conf( item )
        file = open(filename,'a')
        file.write( buffer )
        file.close()
        return True        
            
    def edit_object(self,item, field_name, new_value):
        """
        Modifies a (currently existing) item. Changes are immediate (i.e. there is no commit)
        
        Example Usage: edit_object( item, field_name="host_name", new_value="examplehost.example.com")
        
        THIS FUNCTION IS DEPRECATED. USE item_edit_field() instead
        """
        return self.item_edit_field(item=item, field_name=field_name, new_value=new_value)

    def compareObjects(self, item1, item2):
        """
        Compares two items. Returns true if they are equal"
        """
        keys1 = item1['meta']['defined_attributes'].keys()
        keys2 = item2['meta']['defined_attributes'].keys()
        keys1.sort()
        keys2.sort()
        result = True
        if keys1 != keys2:
            return False
        for key in keys1:
            if key == 'meta': continue
            key1 = item1[key]
            key2 = item2[key]
            # For our purpose, 30 is equal to 30.000
            if key == 'check_interval':
                key1 = int(float(key1))
                key2 = int(float(key2))
            if str(key1) != str(key2):
                result = False
        if result == False: return False
        return True

    def edit_service(self, target_host, service_description, field_name, new_value):
        """
        Edit a service's attributes
        """

        original_object = self.get_service(target_host, service_description)
        if original_object is None:
            raise ParserError("Service not found")
        return self.edit_object( original_object, field_name, new_value)

    def _get_list(self, object, key):
        """
        Return a comma list from an item

        Example:

        _get_list(Foo_object, host_name)
        define service {
            service_description Foo
            host_name            larry,curly,moe
        }

        return
        ['larry','curly','moe']
        """
        if type(object) != type({}):
            raise ParserError("%s is not a dictionary\n" % object)
            # return []
        if not object.has_key(key):
            return []

        return_list = []

        if object[key].find(",") != -1:
            for name in object[key].split(","):
                return_list.append(name)
        else:
            return_list.append(object[key])

        ## Alphabetize
        return_list.sort()

        return return_list
        
    def delete_object(self, object_type, object_name, user_key = None):
        """
        Delete object from configuration files.
        """
        object_key = self._get_key(object_type,user_key)

        k = 'all_%s' % object_type
        if k not in self.data: return None
        for item in self.data[k]:
            if not item.has_key(object_key):
                continue

            ## If the object matches, mark it for deletion
            if item[object_key] == object_name:
                self.data[k].remove(item)
                item['meta']['delete_me'] = True
                item['meta']['needs_commit'] = True
                self.data[k].append(item)

                ## Commit the delete
                self.commit()
                return True

        ## Only make it here if the object isn't found
        return None

    def delete_service(self, service_description, host_name):
        """
        Delete service from configuration
        """
        for item in self.data['all_service']:
            if ('service_description' in item and item['service_description'] == service_description) and (host_name in self._get_active_hosts(item)):
                self.data['all_service'].remove(item)
                item['meta']['delete_me'] = True
                item['meta']['needs_commit'] = True
                self.data['all_service'].append(item)

                return True

    def delete_host(self, object_name, user_key = None):
        """
        Delete a host
        """
        return self.delete_object('host',object_name, user_key = user_key)

    def delete_hostgroup(self, object_name, user_key = None):
        """
        Delete a hostgroup
        """
        return self.delete_object('hostgroup',object_name, user_key = user_key)

    def get_object(self, object_type, object_name, user_key = None):
        """
        Return a complete object dictionary
        """
        object_key = self._get_key(object_type,user_key)

        target_object = None

        #print object_type
        for item in self.data['all_%s' % object_type]:
            ## Skip items without the specified key
            if not item.has_key(object_key):
                continue
            if item[object_key] == object_name:
                target_object = item
            ## This is for multi-key items
        return target_object

    def get_host(self, object_name, user_key = None):
        """
        Return a host object
        """
        return self.get_object('host',object_name, user_key = user_key)

    def get_servicegroup(self, object_name, user_key = None):
        """
        Return a Servicegroup object
        """
        return self.get_object('servicegroup',object_name, user_key = user_key)

    def get_contact(self, object_name, user_key = None):
        """
        Return a Contact object
        """
        return self.get_object('contact',object_name, user_key = user_key)

    def get_contactgroup(self, object_name, user_key = None):
        """
        Return a Contactgroup object
        """
        return self.get_object('contactgroup',object_name, user_key = user_key)

    def get_timeperiod(self, object_name, user_key = None):
        """
        Return a Timeperiod object
        """
        return self.get_object('timeperiod',object_name, user_key = user_key)

    def get_command(self, object_name, user_key = None):
        """
        Return a Command object
        """
        return self.get_object('command',object_name, user_key = user_key)

    def get_hostgroup(self, object_name, user_key = None):
        """
        Return a hostgroup object
        """
        return self.get_object('hostgroup',object_name, user_key = user_key)

    def get_servicedependency(self, object_name, user_key = None):
        """
        Return a servicedependency object
        """
        return self.get_object('servicedependency',object_name, user_key = user_key)

    def get_hostdependency(self, object_name, user_key = None):
        """
        Return a hostdependency object
        """
        return self.get_object('hostdependency',object_name, user_key = user_key)

    def get_service(self, target_host, service_description):
        """
        Return a service object.  This has to be seperate from the 'get_object'
        method, because it requires more than one key
        """
        for item in self.data['all_service']:
            ## Skip service with no service_description
            if not item.has_key('service_description'):
                continue
            ## Skip non-matching services
            if item['service_description'] != service_description:
                continue

            if target_host in self._get_active_hosts(item):
                return item

        return None

    def _append_use(self, source_item, name):
        """
        Append any unused values from 'name' to the dict 'item'
        """
        ## Remove the 'use' key
        if source_item.has_key('use'):
            del source_item['use']
        
        for possible_item in self.pre_object_list:
            if possible_item.has_key('name'):
                ## Start appending to the item
                for k,v in possible_item.iteritems():

                    try:
                        if k == 'use':
                            source_item = self._append_use(source_item, v)
                    except Exception:
                        raise ParserError( "Recursion error on %s %s" % (source_item, v) )

                    ## Only add the item if it doesn't already exist
                    if not source_item.has_key(k):
                        source_item[k] = v
        return source_item

    def _post_parse(self):
        self.item_list = None
        self.item_apply_cache = {} # This is performance tweak used by _apply_template
        for raw_item in self.pre_object_list:
            # Performance tweak, make sure hashmap exists for this object_type
            object_type = raw_item['meta']['object_type']
            if not self.item_apply_cache.has_key( object_type ):
                self.item_apply_cache[ object_type ] = {}
            # Tweak ends
            if raw_item.has_key('use'):
                raw_item = self._apply_template( raw_item )
            self.post_object_list.append(raw_item)
        ## Add the items to the class lists.  
        for list_item in self.post_object_list:
            type_list_name = "all_%s" % list_item['meta']['object_type']
            if not self.data.has_key(type_list_name):
                self.data[type_list_name] = []

            self.data[type_list_name].append(list_item)

    def commit(self):
        """
        Write any changes that have been made to it's appropriate file
        """
        ## Loops through ALL items
        for k in self.data.keys():
            for item in self[k]:

                ## If the object needs committing, commit it!
                if item['meta']['needs_commit']:
                    ## Create file contents as an empty string
                    file_contents = ""

                    ## find any other items that may share this config file
                    extra_items = self._get_items_in_file(item['meta']['filename'])
                    if len(extra_items) > 0:
                        for commit_item in extra_items:
                            ## Ignore files that are already set to be deleted:w
                            if commit_item['meta']['delete_me']:
                                continue
                            ## Make sure we aren't adding this thing twice
                            if item != commit_item:
                                file_contents += self.print_conf(commit_item)

                    ## This is the actual item that needs commiting
                    if not item['meta']['delete_me']:
                        file_contents += self.print_conf(item)

                    ## Write the file
                    f = open(item['meta']['filename'], 'w')
                    f.write(file_contents)
                    f.close()

                    ## Recreate the item entry without the commit flag
                    self.data[k].remove(item)
                    item['meta']['needs_commit'] = None
                    self.data[k].append(item)

    def flag_all_commit(self):
        """
        Flag every item in the configuration to be committed
        This should probably only be used for debugging purposes
        """
        for object_type in self.data.keys():
            for item in self.data[object_type]:
                item['meta']['needs_commit'] = True

    def print_conf(self, item):
        """
        Return a string that can be used in a configuration file
        """
        output = ""
        ## Header, to go on all files
        output += "# Configuration file %s\n" % item['meta']['filename']
        output += "# Edited by PyNag on %s\n" % time.ctime()

        ## Some hostgroup information
        if item['meta'].has_key('hostgroup_list'):
            output += "# Hostgroups: %s\n" % ",".join(item['meta']['hostgroup_list'])

        ## Some hostgroup information
        if item['meta'].has_key('service_list'):
            output += "# Services: %s\n" % ",".join(item['meta']['service_list'])

        ## Some hostgroup information
        if item['meta'].has_key('service_members'):
            output += "# Service Members: %s\n" % ",".join(item['meta']['service_members'])

        if len(item['meta']['template_fields']) != 0:
            output += "# Values from templates:\n"
        for k in item['meta']['template_fields']:
            output += "#\t %-30s %-30s\n" % (k, item[k])
        output += "\n"
        output += "define %s {\n" % item['meta']['object_type']
        for k, v in item.iteritems():
            if k != 'meta':
                if k not in item['meta']['template_fields']:
                    output += "\t %-30s %-30s\n" % (k,v)
        
        output += "}\n\n"
        return output

    def _load_static_file(self, filename=None):
        """Load a general config file (like nagios.cfg) that has key=value config file format. Ignore comments

        Arguments:
          filename -- name of file to parse, if none nagios.cfg will be used

        Returns:
            a [ (key,value), (key,value) ] list
        """
        result = []
        if not filename:
            filename = self.cfg_file
        for line in open(filename).readlines():
            ## Strip out new line characters
            line = line.strip()

            ## Skip blank lines
            if line == "":
                continue

            ## Skip comments
            if line[0] == "#" or line[0] == ';':
                continue
            key, value = line.split("=", 1)
            key = key.strip()
            value = value.strip()
            result.append( (key, value) )
        return result

    def _edit_static_file(self, attribute, new_value, filename=None,old_value=None,append=False):
        """ Modify a general config file (like nagios.cfg) that has a key=value config file format.

        Arguments:
            filename -- Name of config file that will be edited (i.e. nagios.cfg)
            attribute -- name of attribute to edit (i.e. check_external_commands)
            new_value -- new value for the said attribute (i.e. "1"). None deletes the line.
            old_value -- Useful if multiple attributes exist (i.e. cfg_dir) and you want to replace a specific one.
            append -- If true, do not overwrite current setting. Instead append this at the end. Use this with settings that are repeated like cfg_file.
        Examples:
            _edit_static_file(filename='/etc/nagios/nagios.cfg', attribute='check_external_commands', new_value='1')
            _edit_static_file(filename='/etc/nagios/nagios.cfg', attribute='cfg_dir', new_value='/etc/nagios/okconfig', append=True)
        """
        if filename is None:
            filename = self.cfg_file

        # For some specific attributes, append should be implied
        if attribute in ('cfg_file','cfg_dir','broker_module'):
            append = True

        # If/when we make a change, new_line is what will be written
        new_line = '%s=%s\n' % (attribute,new_value)

        # new_value=None means line should be removed
        if new_value is None:
            new_line = ''

        write_buffer = open(filename).readlines()
        is_dirty = False # dirty if we make any changes
        for i,line in enumerate(write_buffer):
            ## Strip out new line characters
            line = line.strip()

            ## Skip blank lines
            if line == "":
                continue

            ## Skip comments
            if line[0] == "#" or line[0] == ';':
                continue
            key, value = line.split("=", 1)
            key = key.strip()
            value = value.strip()

            # If key does not match, we are not interested in this line
            if key != attribute:
                continue

            # If old_value was specified, and it matches, dont have to look any further
            elif value == old_value:
                write_buffer[i] = new_line
                is_dirty = True
                break
            # if current value is the same as new_value, no need to make changes
            elif value == new_value:
                return False
            # Special so cfg_dir matches despite double-slashes, etc
            elif attribute == 'cfg_dir' and os.path.normpath(value) == os.path.normpath(new_value):
                return False
            # We are not appending, and no old value was specified:
            elif append == False and not old_value:
                write_buffer[i] = new_line
                is_dirty = True
                break
        if is_dirty == False:
            # If we get here, it means we read the whole file,
            # and we have not yet made any changes, So we assume
            # We should append to the file
            write_buffer.append(new_line)
            is_dirty = True
        # When we get down here, it is time to write changes to file
        if is_dirty == True:
            open(filename,'w').write(''.join(write_buffer))
            return True
        else:
            return False

    def needs_reload(self):
        """Returns True if Nagios service needs reload of cfg files

        Returns True also if nagios is not running.
        """
        new_timestamps = self.get_timestamps()
        object_cache_file = None
        for k,v in self.maincfg_values:
            if k == 'object_cache_file': object_cache_file = v
        if not object_cache_file:
            return True
        if not os.path.isfile(object_cache_file):
            return True
        object_cache_timestamp = new_timestamps.get(object_cache_file, 0)
        # Reload not needed if no object_cache file
        if object_cache_file is None:
            return False
        for k,v in new_timestamps.items():
            if not v or int(v) > object_cache_file:
                return True
        return False 

    def needs_reparse(self):
        """Returns True if any Nagios configuration file has changed since last parse()"""
        # If Parse has never been run:
        if self.data == {}:
            return True
        new_timestamps = self.get_timestamps()
        if len(new_timestamps) != len( self.timestamps ):
            return True
        for k,v in new_timestamps.items():
            if self.timestamps.get(k, None) != v:
                return True
        return False

    def parse(self):
        """
        Load the nagios.cfg file and parse it up
        """

        # reset
        self.reset()

        self.maincfg_values = self._load_static_file(self.cfg_file)
        
        self.cfg_files = self.get_cfg_files()

        # When parsing config, we will softly fail if permission denied
        # comes on resource files. If later someone tries to get them via
        # get_resource, we will fail hard
        try:
            self._resource_values = self.get_resources()
        except IOError, e:
            self.errors.append( str(e) )
        
        self.timestamps = self.get_timestamps()
        
        ## This loads everything into
        for cfg_file in self.cfg_files:
            self._load_file(cfg_file)

        self._post_parse()

    def get_resource(self, resource_name):
        """ Get a single resource value which can be located in any resource.cfg file

         Arguments:
            resource_name: Name as it appears in resource file (i.e. $USER1$)
        Returns:
            String value of the resource value.
        Raises:
            * KeyError if resource is not found
            * ParserError if resource is not found and you do not have permissions
        """
        resources = self.get_resources()
        for k,v in resources:
            if k == resource_name:
                return v


    def get_timestamps(self):
        """Returns a hash map of all nagios related files and their timestamps"""
        files = {}
        files[self.cfg_file] = None
        for k,v in self.maincfg_values:
            if k in ('resource_file','lock_file','object_cache_file'):
                files[v] = None
        for i in self.get_cfg_files():
            files[i] = None
        # Now lets lets get timestamp of every file
        for k,v in files.items():
            if not os.path.isfile(k): continue
            files[k] = os.stat(k).st_mtime
        return files

    def get_resources(self):
        """Returns a list of every private resources from nagios.cfg"""
        resources = []
        for config_object,config_value in self.maincfg_values:
            if config_object == 'resource_file' and os.path.isfile(config_value):
                resources += self._load_static_file(config_value)
        return resources

    def extended_parse(self):
        """
        This parse is used after the initial parse() command is run.  It is
        only needed if you want extended meta information about hosts or other objects
        """
        ## Do the initial parsing
        self.parse()

        ## First, cycle through the hosts, and append hostgroup information
        index = 0
        for host in self.data['all_host']:
            if host.has_key('register') and host['register'] == '0': continue
            if not host.has_key('host_name'): continue
            if not self.data['all_host'][index]['meta'].has_key('hostgroup_list'):
                self.data['all_host'][index]['meta']['hostgroup_list'] = []

            ## Append any hostgroups that are directly listed in the host definition
            if host.has_key('hostgroups'):
                for hostgroup_name in self._get_list(host, 'hostgroups'):
                    if not self.data['all_host'][index]['meta'].has_key('hostgroup_list'):
                        self.data['all_host'][index]['meta']['hostgroup_list'] = []
                    if hostgroup_name not in self.data['all_host'][index]['meta']['hostgroup_list']:
                        self.data['all_host'][index]['meta']['hostgroup_list'].append(hostgroup_name)

            ## Append any services which reference this host
            service_list = []
            for service in self.data['all_service']:
                if service.has_key('register') and service['register'] == '0': continue
                if not service.has_key('service_description'): continue
                if host['host_name'] in self._get_active_hosts(service):
                    service_list.append(service['service_description'])
            self.data['all_host'][index]['meta']['service_list'] = service_list

            ## Increment count
            index += 1

        ## Loop through all hostgroups, appending them to their respective hosts
        for hostgroup in self.data['all_hostgroup']:
            for member in self._get_list(hostgroup,'members'):
                index = 0
                for host in self.data['all_host']:
                    if not host.has_key('host_name'): continue

                    ## Skip members that do not match
                    if host['host_name'] == member:

                        ## Create the meta var if it doesn' exist
                        if not self.data['all_host'][index]['meta'].has_key('hostgroup_list'):
                            self.data['all_host'][index]['meta']['hostgroup_list'] = []

                        if hostgroup['hostgroup_name'] not in self.data['all_host'][index]['meta']['hostgroup_list']:
                            self.data['all_host'][index]['meta']['hostgroup_list'].append(hostgroup['hostgroup_name'])

                    ## Increment count
                    index += 1

        ## Expand service membership
        index = 0
        for service in self.data['all_service']:

            ## Find a list of hosts to negate from the final list
            self.data['all_service'][index]['meta']['service_members'] = self._get_active_hosts(service)

            ## Increment count
            index += 1

    def _get_active_hosts(self, object):
        """
        Given an object, return a list of active hosts.  This will exclude hosts that ar negated with a "!"
        """
        ## First, generate the negation list
        negate_hosts = []

        ## Hostgroups
        if object.has_key("hostgroup_name"):

            for hostgroup_name in self._get_list(object, 'hostgroup_name'):
                if hostgroup_name[0] == "!":
                    hostgroup_obj = self.get_hostgroup(hostgroup_name[1:])
                    negate_hosts.extend(self._get_list(hostgroup_obj,'members'))

        ## Host Names
        if object.has_key("host_name"):
            for host_name in self._get_list(object, 'host_name'):
                if host_name[0] == "!":
                    negate_hosts.append(host_name[1:])

        ## Now get hosts that are actually listed
        active_hosts = []

        ## Hostgroups
        if object.has_key("hostgroup_name"):

            for hostgroup_name in self._get_list(object, 'hostgroup_name'):
                if hostgroup_name[0] != "!":
                    active_hosts.extend(self._get_list(self.get_hostgroup(hostgroup_name),'members'))

        ## Host Names
        if object.has_key("host_name"):
            for host_name in self._get_list(object, 'host_name'):
                if host_name[0] != "!":
                    active_hosts.append(host_name)

        ## Combine the lists
        return_hosts = []
        for active_host in active_hosts:
            if active_host not in negate_hosts:
                return_hosts.append(active_host)

        return return_hosts

    def get_cfg_dirs(self):
        """
        Return a list of all cfg directories used in this configuration

        Example:
        print get_cfg_dirs()
        ['/etc/nagios/hosts','/etc/nagios/objects',...]
        """
        cfg_dirs = []
        for config_object, config_value in self.maincfg_values:
            if config_object == "cfg_dir":
                cfg_dirs.append(config_value)
        return cfg_dirs

    def get_cfg_files(self):
        """
        Return a list of all cfg files used in this configuration

        Example:
        print get_cfg_files()
        ['/etc/nagios/hosts/host1.cfg','/etc/nagios/hosts/host2.cfg',...]
        """
        cfg_files = []
        for config_object, config_value in self.maincfg_values:
            
            ## Add cfg_file objects to cfg file list
            if config_object == "cfg_file" and os.path.isfile(config_value):
                    cfg_files.append(config_value)

            ## Parse all files in a cfg directory
            if config_object == "cfg_dir":
                directories = []
                raw_file_list = []
                directories.append( config_value )
                # Walk through every subdirectory and add to our list
                while len(directories) > 0:
                    current_directory = directories.pop(0)
                    # Nagios doesnt care if cfg_dir exists or not, so why should we ?
                    if not os.path.isdir( current_directory ): continue
                    list = os.listdir(current_directory)
                    for item in list:
                        # Append full path to file
                        item = "%s" % (os.path.join(current_directory, item.strip()))
                        if os.path.islink( item ):
                            item = os.readlink( item )
                        if os.path.isdir(item):
                            directories.append( item )
                        if raw_file_list.count( item ) < 1:
                            raw_file_list.append( item )
                for raw_file in raw_file_list:
                    if raw_file.endswith('.cfg'):
                        if os.path.exists(raw_file):
                            # Nagios doesnt care if cfg_file exists or not, so we will not throws errors
                            cfg_files.append(raw_file)

        return cfg_files

    def get_object_types(self):
        """ Returns a list of all discovered object types """
        return map(lambda x: re.sub("all_","", x), self.data.keys())

    def cleanup(self):
        """
        This cleans up dead configuration files
        """
        for filename in self.cfg_files:
            if os.path.isfile(filename):
                size = os.stat(filename)[6]
                if size == 0:
                    os.remove(filename)

        return True

    def __setitem__(self, key, item):
        self.data[key] = item

    def __getitem__(self, key):
        return self.data[key]


class status:
    """ Easy way to parse status.dat file from nagios

    After calling parse() contents of status.dat are kept in status.data
    Example usage:
    >>> s = status() # filename autodetected
    >>> s.parse()
    >>> print s.data.keys()
    ['info', 'servicestatus', 'contactstatus', 'hoststatus', 'programstatus']
    >>> for service in s.data.get('servicestatus',[]):
    ...     host_name=service.get('host_name', None)
    ...     description=service.get('service_description',None)
    """

    def __init__(self, filename=None, cfg_file=None ):
        """ Initilize a new instance of status

        Arguments (you only need to provide one of these):
            filename -- path to your status.dat file
            cfg_file -- path to your nagios.cfg file, path to status.dat
              will be looked up in this file
        """
        # If filename is not provided, lets try to discover it from
        # nagios.cfg
        if filename is None:
            c = config(cfg_file=cfg_file)
            for key,value in c._load_static_file():
                if key == "status_file":
                    filename=value

        self.filename = filename
        self.data = None

    def parse(self):
        """ Parses your status.dat file and stores in a dictionary under self.data

        Returns:
            None
        Raises:
            ParserError -- if problem arises while reading status.dat
            ParserError -- if status.dat is not found
            IOError -- if status.dat cannot be read
        """
        self.data = {}
        status = {} # Holds all attributes of a single item
        key = None # if within definition, store everything before =
        value = None # if within definition, store everything after =
        lines = open(self.filename, 'rb').readlines()
        for line_num,line in enumerate( lines ):

            ## Cleanup and line skips
            line = line.strip()
            if line == "":
                pass
            elif line[0] == "#" or line[0] == ';':
                pass
            elif line.find("{") != -1:
                status = {}
                status['meta'] = {}
                status['meta']['type'] = line.split("{")[0].strip()
            elif line.find("}") != -1:
                # Status definition has finished, lets add it to
                # self.data
                if status['meta']['type'] not in self.data:
                    self.data[status['meta']['type']] = []
                self.data[status['meta']['type']].append(status)
            else:
                tmp = line.split("=", 1)
                if len(tmp) == 2:
                    (key, value) = line.split("=", 1)
                    status[key] = value
                elif key == "long_plugin_output":
                    # special hack for long_output support. We get here if:
                    # * line does not contain {
                    # * line does not contain }
                    # * line does not contain =
                    # * last line parsed started with long_plugin_output=
                    status[key] += "\n" + line
                else:
                    raise ParserError("Error on %s:%s: Could not parse line: %s" % (self.filename,line_num,line))
    def get_contactstatus(self, contact_name):
        """ Returns a dictionary derived from status.dat for one particular contact

        Returns:
            dict
        Raises:
            ValueError if object is not found
        >>> s = status()
        >>> s.get_contactstatus(contact_name='invalid_contact')
        ValueError('invalid_contact',)
        >>> first_contact = s.data['contactstatus'][0]['contact_name']
        >>> s.get_contactstatus(first_contact)['contact_name'] == first_contact
        True
        """
        if self.data is None:
            self.parse()
        for i in self.data['contactstatus']:
            if i.get('contact_name') == contact_name:
                return i
        return ValueError(contact_name)
    def get_hoststatus(self, host_name):
        """ Returns a dictionary derived from status.dat for one particular contact

        Returns:
            dict
        Raises:
            ValueError if object is not found
        """
        if self.data is None:
            self.parse()
        for i in self.data['hoststatus']:
            if i.get('host_name') == host_name:
                return i
        return ValueError(host_name)
    def get_servicestatus(self, host_name, service_description):
        """ Returns a dictionary derived from status.dat for one particular service
        Returns:
            dict
        Raises:
            ValueError if object is not found
        """
        if self.data is None:
            self.parse()
        for i in self.data['servicestatus']:
            if i.get('host_name') == host_name:
                if i.get('service_description') == service_description:
                    return i
        return ValueError(host_name, service_description)


    def __setitem__(self, key, item):
        self.data[key] = item

    def __getitem__(self, key):
        return self.data[key]


class object_cache(config):
    """ Loads the configuration as it appears in objects.cache file """
    def get_cfg_files(self):
        for k,v in self.maincfg_values:
            if k == 'object_cache_file': return [ v ]


class ParserError(Exception):
    """ ParserError is used for errors that the Parser has when parsing config.

    Typical usecase when there is a critical error while trying to read configuration.
    """
    def __init__(self, message, item=None):
        self.message = message
        if item is None: return
        self.item = item
        self.filename = item['meta']['filename']

    def __str__(self):
        return repr(self.message)


if __name__ == '__main__':
    c = config()
    print c._edit_static_file('test','new_value')




# -*- coding: utf-8 -*-
#
# pynag - Python Nagios plug-in and configuration environment
# Copyright (C) 2010 Drew Stinnet
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys
import os
import re

from pynag.Utils import PynagError

class daemon:
    """
    Control the nagios daemon through python
    """

    def __init__(self, nagios_bin = "/usr/bin/nagios", nagios_cfg = "/etc/nagios/nagios.cfg", nagios_init = "/etc/init.d/nagios"):

        self.nagios_bin = nagios_bin
        self.nagios_cfg = nagios_cfg
        self.nagios_init = nagios_init

    def verify_config(self):
        """
        Run nagios -v config_file to verify that the conf is working
        """

        cmd = "%s -v %s" % (self.nagios_bin, self.nagios_cfg)

        result = os.system(cmd)

        if result == 0:
            return True
        else:
            return None

    def restart(self):
        cmd = "%s restart" % self.nagios_init

        return os.system(cmd)
    def status(self):
        cmd = "%s status" % self.nagios_init

        return os.system(cmd)
    def reload(self):
        cmd = "%s reload" % self.nagios_init

        return os.system(cmd)



# -*- coding: utf-8 -*-
#
# pynag - Python Nagios plug-in and configuration environment
# Copyright (C) 2010 Drew Stinnet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


""" Misc utility classes and helper functions for pynag

This module contains misc classes and conveninence functions
that are used throughout the pynag library.

"""
import subprocess


class PynagError(Exception):
    """ The default pynag exception.

    Exceptions raised within the pynag library should aim
    to inherit this one.

    """



def runCommand(command, raise_error_on_fail=False):
    """ Run command from the shell prompt. Wrapper around subprocess.

     Arguments:
         command: string containing the command line to run
         raise_error_on_fail: Raise PynagError if returncode >0
     Returns:
         stdout/stderr of the command run
     Raises:
         PynagError if returncode > 0
     """
    proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,stderr=subprocess.PIPE,)
    stdout, stderr = proc.communicate('through stdin to stdout')
    result = proc.returncode,stdout,stderr
    if proc.returncode > 0 and raise_error_on_fail==True:
        error_string = "* Could not run command (return code= %s)\n" % proc.returncode
        error_string += "* Error was:\n%s\n" % (stderr.strip())
        error_string += "* Command was:\n%s\n" % command
        error_string += "* Output was:\n%s\n" % (stdout.strip())
        if proc.returncode == 127: # File not found, lets print path
            path=getenv("PATH")
            error_string += "Check if y/our path is correct: %s" % path
        raise PynagError( error_string )
    else:
        return result


class GitRepo(object):
    def __init__(self, directory):
        """
        Python Wrapper around Git command line.


        """

        self.directory = directory

        # Who made the change
        self.modified_by = modified_by

        # Which program did the change
        self.source = source

        # Every string in self.messages indicated a line in the eventual commit message
        self.messages = []

        self.ignore_errors = ignore_errors
        if auto_init:
            try:
                self._run_command('git status --short')
            except EventHandlerError, e:
                if e.errorcode == 128:
                    self._git_init()
            #self._run_command('git status --short')

        self._update_author()
    def _update_author(self):
        """ Updates environment variables GIT_AUTHOR_NAME and EMAIL

        Returns: None
        """
        environ['GIT_AUTHOR_NAME'] = self.modified_by
        environ['GIT_AUTHOR_EMAIL'] = "%s@%s" % (self.source, node())
    def _run_command(self, command):
        """ Run a specified command from the command line. Return stdout """
        import subprocess
        import os
        cwd = self.gitdir
        proc = subprocess.Popen(command, cwd=cwd, shell=True, stdout=subprocess.PIPE,stderr=subprocess.PIPE,)
        stdout, stderr = proc.communicate('through stdin to stdout')
        returncode = proc.returncode
        if returncode > 0 and self.ignore_errors == False:
            errorstring = "Command '%s' returned exit status %s.\n stdout: %s \n stderr: %s\n Current user: %s"
            errorstring = errorstring % (command, returncode, stdout, stderr,getuser())
            raise EventHandlerError( errorstring, errorcode=returncode, errorstring=stderr )
        return stdout
    def is_commited(self):
        """ Returns True if all files in git repo are fully commited """
        return self.get_uncommited_files() == 0
    def get_uncommited_files(self):
        """ Returns a list of files that are have unstaged changes """
        output = self._run_command("git status --porcelain")
        result = []
        for line in output.split('\n'):
            line = line.split()
            if len(line) < 2:
                continue
            result.append( {'status':line[0], 'filename': line[1]} )
        return result
    def init(self):
        """ Initilizes a new git repo (i.e. run "git init") """
        self._update_author()
        command = "git init"
        self._run_command("git init")
        self._run_command("git add .")
        self._run_command("git commit -a -m 'Initial Commit'")
    def _git_add(self, filename):
        """ Wrapper around git add command """
        self._update_author()
        directory = dirname(filename)
        command= "git add '%s'" % filename
        return self._run_command(command)
    def _git_commit(self, filename, message, filelist=[]):
        """ Wrapper around git commit command """
        self._update_author()
        # Lets strip out any single quotes from the message:
        message = message.replace("'",'"')
        if len(filelist) > 0:
            filename = "' '".join(filelist)
        command = "git commit '%s' -m '%s'" % (filename, message)
        return self._run_command(command=command)
    def pre_save(self, object_definition, message):
        """ Commits object_definition.get_filename() if it has any changes """
        filename = object_definition.get_filename()
        if self._is_dirty(filename):
            self._git_add(filename)
            self._git_commit(filename,
                message="External changes commited in %s '%s'" %
                        (object_definition.object_type, object_definition.get_shortname()))
    def save(self, object_definition, message):
        filename = object_definition.get_filename()
        if len(self.messages) > 0:
            message = [message, '\n'] + self.messages
            message = '\n'.join(message)
        self._git_add(filename)
        if self._is_dirty(filename):
            self._git_commit(filename, message)
        self.messages = []
    def _is_dirty(self,filename):
        """ Returns True if filename needs to be committed to git """
        command = "git status --porcelain '%s'" % filename
        output = self._run_command(command)
        # Return True if there is any output
        return len(output) > 0
    def write(self, object_definition, message):
        # When write is called ( something was written to file )
        # We will log it in a buffer, and commit when save() is called.
        self.messages.append( " * %s" % message )


class PerfData(object):
    """ Data Structure for a nagios perfdata string with multiple perfdata metric

    Example string:
    >>> perf = PerfData("load1=10 load2=10 load3=20")
    >>> for i in perf.metrics:
    >>>     print i.label, i.value
    """
    def __init__(self, perfdatastring):
        """ >>> perf = PerfData("load1=10 load2=10 load3=20") """
        import shlex
        perfdata = shlex.split(perfdatastring)
        self.metrics = []
        self.invalid_metrics = []
        for metric in perfdata:
            try:
                self.add_perfdatametric( PerfDataMetric(metric) )
            except Exception:
                self.invalid_metrics.append( metric )
    def is_valid(self):
        """ Returns True if the every metric in the string is valid """
        for i in self.metrics:
            if not i.is_valid():
                return False
    def add_perfdatametric(self, perfdatastring="", label="",value="",warn="",crit="",min="",max="",uom=""):
        """ Add a new perfdatametric to existing list of metrics.

         Example:
         >>> s.add_perfdatametric("a=1")
         >>> s.add_perfdatametric(label="utilization",value="10",uom="%")
        """
        metric=PerfDataMetric(perfdatastring=perfdatastring, label=label,value=value,warn=warn,crit=crit,min=min,max=max,uom=uom)
        self.metrics.append(  metric )

class PerfDataMetric(object):
    """ Data structure for one single Nagios Perfdata Metric """
    label = ""
    value = ""
    warn = ""
    crit = ""
    min = ""
    max = ""
    uom = ""
    status = "ok"
    def __repr__(self):
        return "'%s'=%s%s;%s;%s;%s;%s" % (
            self.label,
            self.value,
            self.uom,
            self.warn,
            self.crit,
            self.min,
            self.max,
            )
    def __str__(self):
        return """
            label: %s
            value: %s %s
            warning: %s
            critical: %s
            min: %s
            max: %s
            """ % (
            self.label,
            self.value,
            self.uom,
            self.warn,
            self.crit,
            self.min,
            self.max,
            )

    def __init__(self, perfdatastring="", label="",value="",warn="",crit="",min="",max="",uom=""):
        """
        >>> p = PerfData(perfdatastring="size=10M;20M;;;"
        >>> print p.label
        size
        >>> print p.value
        >>> 10
        print p.value_unit
        >>> M
        """
        self.label = label
        self.value = value
        self.warn = warn
        self.crit = crit
        self.min = min
        self.max = max
        self.uom = uom

        # If label is single quoted, there might be any symbol in the label
        # including other single quotes and the = sign. Therefore, we take special precautions if it is so
        if perfdatastring.startswith("'"):
            tmp = perfdatastring.split("'")
            everything_but_label = tmp.pop()
            tmp.pop(0)
            label = "'".join(tmp)
        else:
            label, everything_but_label = perfdatastring.split('=', 1)
        self.label = label

        # Next split string into value;warning;critical;min;max
        tmp = everything_but_label.split(';')
        if len(tmp) > 0:
            val = tmp.pop(0).strip('=')
            self.value, self.uom = self.split_value_and_uom(val)
        if len(tmp) > 0:
            self.warn = tmp.pop(0)
        if len(tmp) > 0:
            self.crit = tmp.pop(0)
        if len(tmp) > 0:
            self.min = tmp.pop(0)
        if len(tmp) > 0:
            self.max = tmp.pop(0)
        self.value = float(self.value)
    def get_status(self):
        """ Return nagios-style exit code (int 0-3) by comparing

          self.value with self.warn and self.crit
        """
        status = check_threshold(self.value, warning=self.warn, critical=self.crit)
        return status


    def is_valid(self):
        """ Returns True if all Performance data is valid. Otherwise False """
        try:
            self.value == '' or float(self.value)
        except ValueError:
            return False
        try:
            self.min == '' or float(self.min)
        except ValueError:
            return False
        try:
            self.max == '' or float(self.max)
        except ValueError:
            return False
        if self.label.find(' ') > -1 and not self.label.startswith("'") and not self.label.endswith("'"):
            return False


    def split_value_and_uom(self, value):
        """ get value="10M" and return (10,"M")

        >>> split_value_and_uom( "10" )
        (10, '')
        >>> split_value_and_uom( "10c" )
        (10, 'c')
        >>> split_value_and_uom( "10B" )
        (10, 'B')
        >>> split_value_and_uom( "10MB" )
        (10, 'MB')
        >>> split_value_and_uom( "10KB" )
        (10, 'KB')
        >>> split_value_and_uom( "10TB" )
        (10, 'TB')
        >>> split_value_and_uom( "10%" )
        (10, '%')
        >>> split_value_and_uom( "10s" )
        (10, 's')
        >>> split_value_and_uom( "10us" )
        (10, 'us')
        >>> split_value_and_uom( "10ms" )
        (10, 'ms')
        """
        tmp = re.findall(r"([-]*[\d.]*\d+)(.*)", value)
        if len(tmp) == 0:
            return '',''
        return tmp[0]



#!/usr/bin/python

## setup.py ###
from distutils.core import setup
from pynag import __version__

NAME = "pynag"
SHORT_DESC = "Python Modules for Nagios plugins and configuration" 
LONG_DESC = """
Contains python modules for pragmatically handling configuration
file maintenance and plugin development.
"""

if __name__ == "__main__":
    manpath        = "share/man/man1/"
    etcpath = "/etc/%s" % NAME
    etcmodpath    = "/etc/%s/modules" % NAME
    initpath    = "/etc/init.d/"
    logpath        = "/var/log/%s/" % NAME
    varpath        = "/var/lib/%s/" % NAME
    rotpath        = "/etc/logrotate.d"
    setup(
        name='%s' % NAME,
        version = __version__,
        author='Drew Stinnett',
        description = SHORT_DESC,
        long_description = LONG_DESC,
        author_email='drew@drewlink.com',
        url='http://code.google.com/p/pynag/',
        license='GPLv2',
        scripts = [
            'scripts/pynag'
        ],
        packages = [
            'pynag',
            'pynag.Model',
            'pynag.Model.EventHandlers',
            'pynag.Plugins',
            'pynag.Parsers',
            'pynag.Control',
            'pynag.Utils',
        ],
          data_files = [(manpath, [
        'docs/pynag.1.gz',
        ]),
        ],
    )



# -*- coding: utf-8 -*-
#
# pynag - Python Nagios plug-in and configuration environment
# Copyright (C) 2010 Drew Stinnet
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Python Nagios extensions
"""

import sys
import os
import re
from platform import node
from optparse import OptionParser


# Map the return codes
OK = 0
WARNING = 1
CRITICAL = 2
UNKNOWN = 3

class simple:
    """
    Nagios plugin helper library based on Nagios::Plugin

    Sample usage

    from pynag.Plugins import WARNING, CRITICAL, OK, UNKNOWN, simple as Plugin

    # Create plugin object
    np = Plugin()
    # Add arguments
    np.add_arg("d", "disk")
    # Do activate plugin
    np.activate()
    ... check stuff, np['disk'] to address variable assigned above...
    # Add a status message and severity
    np.add_message( WARNING, "Disk nearing capacity" )
    # Get parsed code and messages
    (code, message) = np.check_messages()
    # Return information and exit
    nagios_exit(code, message)
    """

    def __init__(self, shortname = None, version = None, blurb = None, extra = None, url = None, license = None, plugin = None, timeout = 15, must_threshold = True):

        ## this is the custom parser
        self.extra_list_optional = []
        self.extra_list_required = []

        ## Set the option parser stuff here
        self.parser = OptionParser()

        ## Variables we'll get later
        self.opts = None
        self.must_threshold = must_threshold
        self.data = {}
        self.data['perfdata'] = []
        self.data['messages'] = { OK:[], WARNING:[], CRITICAL:[], UNKNOWN:[] }
        self.data['threshhold'] = None

        ## Error mappings, for easy access
        self.errors = { "OK":0, "WARNING":1, "CRITICAL":2, "UNKNOWN":3, }
        self.status_text = { 0:"OK", 1:"WARNING", 2:"CRITICAL", 3:"UNKNOWN", }

        ## Shortname creation
        if not shortname:
            self.data['shortname'] = os.path.basename("%s" % sys.argv[0])
        else:
            self.data['shortname'] = shortname

    def add_arg(self, spec_abbr, spec, help_text, required=1, action="store"):
        """
        Add an argument to be handled by the option parser.  By default, the arg is not required.
        
        required = optional parameter
        action = [store, append, store_true]
        """
        self.parser.add_option("-%s" % spec_abbr, "--%s" % spec, dest="%s" % spec, help=help_text, metavar="%s" % spec.upper(), action=action)
        if required:
            self.extra_list_required.append(spec)
        else:
            self.extra_list_optional.append(spec)

    def activate(self):
        """
        Parse out all command line options and get ready to process the plugin.  This should be run after argument preps
        """
        timeout = None
        verbose = 0

        self.parser.add_option("-v", "--verbose", dest="verbose", help="Verbosity Level", metavar="VERBOSE", default=0)
        self.parser.add_option("-H", "--host", dest="host", help="Target Host", metavar="HOST")
        self.parser.add_option("-t", "--timeout", dest="timeout", help="Connection Timeout", metavar="TIMEOUT")

        if self.must_threshold == True:
            self.parser.add_option("-c", "--critical", dest="critical", help="Critical Threshhold", metavar="CRITICAL")
            self.parser.add_option("-w", "--warning", dest="warning", help="Warn Threshhold", metavar="WARNING")

        (options, args) = self.parser.parse_args()

        ## Set verbosity level
        if int(options.verbose) in (0, 1, 2, 3):
            self.data['verbosity'] = options.verbose
        else:
            self.data['verbosity'] = 0

        ## Ensure the hostname is set
        if options.host:
            self.data['host'] = options.host

        ## Set timeout
        if options.timeout:
            self.data['timeout'] = options.timeout
        else:
            self.data['timeout'] = timeout

        if self.must_threshold == True and not options.critical and not options.warning:
            self.parser.error("You must provide a WARNING and/or CRITICAL value")

        ## Set Critical
        if options.critical:
            self.data['critical'] = options.critical
        else: self.data['critical'] = None

        ## Set Warn
        if options.warning:
            self.data['warning'] = options.warning
        else:
            self.data['warning'] = None

        ## Ensurethat the extra items are provided
        for extra_item in self.extra_list_required:
            if not options.__dict__[extra_item]:
                self.parser.error("option '%s' is required" % extra_item)


        ## Put the remaining values into the data dictionary
        for key,value in options.__dict__.items():
            if key in (self.extra_list_required + self.extra_list_optional):
                self.data[key] = value

    def add_perfdata(self, label , value , uom = None, warn = None, crit = None, minimum = None, maximum = None):
        """
        Append perfdata string to the end of the message
        """

        # Append perfdata (assume multiple)
        self.data['perfdata'].append({ 'label' : label, 'value' : value, 'uom' : uom, 
            'warn' : warn, 'crit' : crit, 'min' : minimum, 'max' : maximum})

    def check_range(self, value):
        """
        Check if a value is within a given range.  This should replace change_threshold eventually. Exits with appropriate exit code given the range.

        Taken from:  http://nagiosplug.sourceforge.net/developer-guidelines.html
        Range definition
    
        Generate an alert if x...
        10        < 0 or > 10, (outside the range of {0 .. 10})
        10:        < 10, (outside {10 .. #})
        ~:10    > 10, (outside the range of {-# .. 10})
        10:20    < 10 or > 20, (outside the range of {10 .. 20})
        @10:20    # 10 and # 20, (inside the range of {10 .. 20})
        """
        critical = self.data['critical']
        warning = self.data['warning']
        self.hr_range = ""

        if critical and self._range_checker(value, critical):
            self.add_message(CRITICAL,"%s is within critical range: %s" % (value, critical))
        elif warning and self._range_checker(value, warning):
            self.add_message(WARNING,"%s is within warning range: %s" % (value, warning))
        else:
            self.add_message(OK,"%s is outside warning=%s and critical=%s" % (value, warning, critical))

        # Get all messages appended and exit code
        (code, message) = self.check_messages()

        # Exit with appropriate exit status and message
        self.nagios_exit(code, message)

    def _range_checker(self, value, range_threshold):
        """ deprecated. Use pynag.Plugins.check_range() """
        return check_range(value=value, range_threshold=range_threshold)

    def send_nsca(self, code, message, ncsahost, hostname=node(), service=None):
        """
        Send data via send_nsca for passive service checks
        """
    
        # Execute send_nsca
        from popen2 import Popen3
        command = "send_nsca -H %s" % ncsahost
        p = Popen3(command,  capturestderr=True)

        # Service check
        if service:
            print >>p.tochild, "%s    %s    %s    %s %s" % (hostname, service, code, message, self.perfdata_string())
        # Host check, omit service_description
        else:
            print >>p.tochild, "%s    %s    %s %s" % (hostname, code, message, self.perfdata_string())

        # Send eof
        # TODO, support multiple statuses ?
        p.tochild.close()

        # Save output incase we have an error
        nsca_output = ''
        for line in p.fromchild.readlines():
            nsca_output += line

        # Wait for send_nsca to exit
        returncode = p.wait()
        returncode = os.WEXITSTATUS( returncode) 

        # Problem with running nsca
        if returncode != 0:
            if returncode == 127:
                raise Exception("Could not find send_nsca in path")
            else:
                raise Exception("returncode: %i\n%s" % (returncode, nsca_output))

        return 0

    def nagios_exit(self, code_text, message):
        """
        Exit with exit_code, message, and optionally perfdata
        """

        # Change text based codes to int
        code = self.code_string2int(code_text)

        ## This should be one line (or more in nagios 3)
        print "%s: %s %s" % (self.status_text[code], message, self.perfdata_string())
        sys.exit(code)

    def perfdata_string(self):

        ## Append perfdata to the message, if perfdata exists
        if self.data['perfdata']:
            append = '|'
        else:
            append = ''

        for pd in self.data['perfdata']:
            append += " '%s'=%s%s;%s;%s;%s;%s" % (
                pd['label'],
                pd['value'],
                pd['uom'] or '',
                pd['warn'] or '',
                pd['crit'] or '',
                pd['min'] or '',
                pd['max'] or '')

        return append

    def add_message( self, code, message ):
        """
        Add a message with code to the object. May be called
        multiple times.  The messages added are checked by check_messages,
        following.

        Only CRITICAL, WARNING, OK and UNKNOWN are accepted as valid codes.
        """
        # Change text based codes to int
        code = self.code_string2int(code)

        self.data['messages'][code].append( message )

    def check_messages( self, joinstr = " ", joinallstr = None ):
        """
        Check the current set of messages and return an appropriate nagios
        return code and/or a result message. In scalar context, returns
        only a return code; in list context returns both a return code and
        an output message, suitable for passing directly to nagios_exit()

        joinstr = string
            A string used to join the relevant array to generate the
            message string returned in list context i.e. if the 'critical'
            array is non-empty, check_messages would return:
                joinstr.join(critical)

        joinallstr = string
            By default, only one set of messages are joined and returned in
            the result message i.e. if the result is CRITICAL, only the
            'critical' messages are included in the result; if WARNING,
            only the 'warning' messages are included; if OK, the 'ok'
            messages are included (if supplied) i.e. the default is to
            return an 'errors-only' type message.

            If joinallstr is supplied, however, it will be used as a string
            to join the resultant critical, warning, and ok messages
            together i.e.  all messages are joined and returned.
        """
        # Check for messages in unknown, critical, warning, ok to determine
        # code
        keys = self.data['messages'].keys()
        keys.sort(reverse=True)
        code = UNKNOWN
        for code in keys:
            if len(self.data['messages'][code]):
                break

        # Create the relevant message for the most severe code
        if joinallstr is None:
            message = joinstr.join(self.data['messages'][code])
        # Join all strings whether OK, WARN...
        else:
            message = ""
            for c in keys:
                if len(self.data['messages'][c]):
                    message += joinallstr.join(self.data['messages'][c]) + joinallstr

        return code, message.rstrip(joinallstr)

    def code_string2int( self, code_text ):
        """
        Changes CRITICAL, WARNING, OK and UNKNOWN code_text to integer
        representation for use within add_message() and nagios_exit()
        """

        # If code_text is a string, convert to the int
        if str(type(code_text)) == "<type 'str'>":
            code = self.errors[code_text]
        else:
            code = code_text

        return code

    def __setitem__(self, key, item):
        self.data[key] = item

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        else:
            return None

def check_threshold(value, warning=None, critical=None):
    """ Checks value against warning/critical and returns Nagios exit code.

    Format of range_threshold is according to:
    http://nagiosplug.sourceforge.net/developer-guidelines.html#THRESHOLDFORMAT

    Returns (in order of appearance):
        UNKNOWN int(3)  -- On errors or bad input
        CRITICAL int(2) -- if value is within critical threshold
        WARNING int(1)  -- If value is within warning threshold
        OK int(0)       -- If value is outside both tresholds
    Arguments:
        value -- value to check
        warning -- warning range
        critical -- critical range

    # Example Usage:
    >>> check_threshold(88, warning="90:", critical="95:")
    0
    >>> check_threshold(92, warning="90:", critical="95:")
    1
    >>> check_threshold(96, warning="90:", critical="95:")
    2
    """
    if critical and not check_range(value, critical):
        return CRITICAL
    elif warning and not check_range(value, warning):
        return WARNING
    else:
        return OK


def check_range(value, range_threshold=None):
    """ Returns True if value is within range_threshold.

    Format of range_threshold is according to:
    http://nagiosplug.sourceforge.net/developer-guidelines.html#THRESHOLDFORMAT

    Arguments:
        value -- Numerical value to check (i.e. 70 )
        range -- Range to compare against (i.e. 0:90 )
    Returns:
        True  -- If value is inside the range
        False -- If value is outside the range (alert if this happens)

    Summary from plugin developer guidelines:
    ---------------------------------------------------------
    x       Generate an alert if x...
    ---------------------------------------------------------
    10  	< 0 or > 10, (outside the range of {0 .. 10})
    10:     < 10, (outside {10 .. ∞})
    ~:10    > 10, (outside the range of {-∞ .. 10})
    10:20   < 10 or > 20, (outside the range of {10 .. 20})
    @10:20  ≥ 10 and ≤ 20, (inside the range of {10 .. 20})
    10      < 0 or > 10, (outside the range of {0 .. 10})
    ---------------------------------------------------------


    # Example runs for doctest, False should mean alert
    >>> check_range(78, "90") # Example disk is 78% full, threshold is 90
    True
    >>> check_range(5, 10) # Everything between 0 and 10 is True
    True
    >>> check_range(0, 10) # Everything between 0 and 10 is True
    True
    >>> check_range(10, 10) # Everything between 0 and 10 is True
    True
    >>> check_range(11, 10) # Everything between 0 and 10 is True
    False
    >>> check_range(-1, 10) # Everything between 0 and 10 is True
    False
    >>> check_range(-1, "~:10") # Everything Below 10
    True
    >>> check_range(11, "10:") # Everything above 10 is True
    True
    >>> check_range(1, "10:") # Everything above 10 is True
    False
    >>> check_range(0, "5:10") # Everything between 5 and 10 is True
    False
    >>> check_range(0, "@5:10") # Everything outside 5:10 is True
    True
    """

    # if no range_threshold is provided, assume everything is ok
    if not range_threshold:
        range_threshold='~:'
    range_threshold = str(range_threshold)
    # If range starts with @, then we do the opposite
    if range_threshold[0] == '@':
        return not check_range(value, range_threshold[1:])

    value = float(value)
    if range_threshold.find(':') > -1:
        (start,end) = (range_threshold.split(':', 1))
    # we get here if ":" was not provided in range_threshold
    else:
        start = ''
        end = range_threshold
    # assume infinity if start is not provided
    if start == '~':
        start = None
    # assume start=0 if start is not provided
    if start == '':
        start = 0
    # assume infinity if end is not provided
    if end == '':
        end = None
    # start is defined and value is lower than start
    if start is not None and float(value) < float(start):
        return False
    if end is not None and float(value) > float(end):
        return False
    return True
